# src/svy/regression/_results.py
"""
Result containers for GLM.
"""

from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Any

import msgspec
import numpy as np
import polars as pl

from msgspec import field

from svy.core.containers import FDist, TDist
from svy.utils.formats import _fmt_fixed, _fmt_p, _fmt_smart


if TYPE_CHECKING:
    pass

log = logging.getLogger(__name__)


# =============================================================================
# UI Integration
# =============================================================================


def _ui_module():
    """Try to import central UI module (optional)."""
    try:
        from svy.ui import printing as _p

        return _p
    except Exception:
        return None


def _resolve_width_for(obj, default: int = 88) -> int:
    """Robust width resolution delegating to central UI."""
    P = _ui_module()
    if P:
        try:
            return int(P.resolve_width(owner=obj, default=default))
        except Exception:
            pass
    return default


# =============================================================================
# Result Structs
# =============================================================================


class GLMCoef(msgspec.Struct, frozen=True):
    """A single coefficient in a regression table."""

    term: str
    est: float
    se: float
    lci: float
    uci: float
    wald: TDist | None = None
    wald_adj: TDist | None = None

    def to_dict(self) -> dict[str, Any]:
        return msgspec.to_builtins(self)


class GLMStats(msgspec.Struct, frozen=True):
    """Model-level goodness-of-fit statistics."""

    n: int
    wald: FDist
    wald_adj: FDist
    scale: float
    deviance: float
    aic: float | None = None
    bic: float | None = None
    r_squared: float | None = None
    r_squared_adj: float | None = None
    iterations: int | None = None

    def to_dict(self) -> dict[str, Any]:
        return msgspec.to_builtins(self)


class GLMFit(msgspec.Struct, frozen=True):
    """Immutable container for fitted GLM results."""

    y: str
    family: str
    link: str
    stats: GLMStats
    coefs: list[GLMCoef] = field(default_factory=list)
    cov_matrix: np.ndarray | None = None
    term_info: dict | None = None  # Encoding info for prediction
    feature_names: list[str] = field(default_factory=list)  # Original feature order

    def to_dict(self) -> dict[str, Any]:
        d = msgspec.to_builtins(self)
        # Don't serialize large matrices
        d.pop("cov_matrix", None)
        d.pop("term_info", None)
        return d

    def to_polars(self) -> pl.DataFrame:
        """Convert coefficients to DataFrame."""
        data = []
        for c in self.coefs:
            row: dict[str, Any] = {
                "term": c.term,
                "estimate": c.est,
                "std_err": c.se,
                "conf_low": c.lci,
                "conf_high": c.uci,
            }
            if c.wald:
                row.update(
                    {
                        "statistic": c.wald.value,
                        "p_value": c.wald.p_value,
                        "df": c.wald.df,
                    }
                )
            else:
                row.update({"statistic": None, "p_value": None, "df": None})

            if c.wald_adj:
                row["adj_statistic"] = c.wald_adj.value
                row["adj_p_value"] = c.wald_adj.p_value
                row["adj_df"] = c.wald_adj.df

            data.append(row)
        return pl.DataFrame(data)

    def __rich_console__(self, console, options):
        from rich import box
        from rich.console import Group
        from rich.table import Table as RTable
        from rich.text import Text

        if not self.coefs:
            yield Text("Empty GLM Model", style="red")
            return

        # 1. Stats Grid
        st = self.stats
        df_resid = st.wald.df_den if st.wald else "-"

        stats_rows = [
            ("Observations", str(st.n), "AIC", _fmt_smart(st.aic)),
            ("Df Residuals", str(df_resid), "BIC", _fmt_smart(st.bic)),
            ("Deviance", _fmt_smart(st.deviance), "Scale", _fmt_smart(st.scale)),
        ]

        if st.r_squared is not None:
            stats_rows.append(
                (
                    "R-squared",
                    _fmt_fixed(st.r_squared),
                    "R-sq (adj)",
                    _fmt_fixed(st.r_squared_adj),
                )
            )

        if st.iterations:
            stats_rows.append(("", "", "Iterations", str(st.iterations)))

        if st.wald_adj:
            stats_rows.append(
                (
                    "F-stat (adj)",
                    _fmt_fixed(st.wald_adj.value),
                    "Prob (F-adj)",
                    _fmt_p(st.wald_adj.p_value),
                )
            )

        stats_grid = RTable.grid(padding=(0, 2))
        stats_grid.add_column(style="bold")
        stats_grid.add_column(justify="right")
        stats_grid.add_column(style="bold")
        stats_grid.add_column(justify="right")

        for r in stats_rows:
            stats_grid.add_row(*r)

        # 2. Coefficients Table
        coef_tbl = RTable(
            show_header=True,
            header_style="bold",
            box=box.SIMPLE_HEAVY,
            show_edge=True,
            show_lines=False,
            pad_edge=False,
            expand=False,
        )

        cols = [
            ("Term", "left"),
            ("Coef.", "right"),
            ("Std.Err.", "right"),
            ("t", "right"),
            ("P>|t|", "right"),
            ("[0.025", "right"),
            ("0.975]", "right"),
        ]
        for name, justify in cols:
            coef_tbl.add_column(name, justify=justify)  # type: ignore[arg-type]

        for row in self.coefs:
            t_val = row.wald.value if row.wald else 0.0
            p_val = row.wald.p_value if row.wald else 1.0

            p_style = "bold red" if p_val < 0.05 else ""
            coef_tbl.add_row(
                row.term,
                _fmt_fixed(row.est),
                _fmt_fixed(row.se),
                _fmt_fixed(t_val),
                Text(_fmt_p(p_val), style=p_style),
                _fmt_fixed(row.lci),
                _fmt_fixed(row.uci),
            )

        # 3. Assemble Content
        content = Group(
            Text(f"Modeling: {self.y}", style="dim"),
            Text(""),
            stats_grid,
            Text(""),
            coef_tbl,
        )

        # 4. Wrap in Panel via Central UI
        P = _ui_module()
        title_text = f"GLM: {self.family} ({self.link})"

        if P and hasattr(P, "make_panel"):
            yield P.make_panel([content], title=title_text, kind="estimate")
        else:
            from rich.panel import Panel  # always available inside __rich_console__

            yield Panel(
                content, title=title_text, title_align="left", border_style="cyan", padding=(0, 1)
            )

    def __plain_str__(self) -> str:
        """Plain-text fallback when rich is not installed. Never calls str(self)."""
        lines = [
            f"GLM: {self.family} ({self.link})",
            f"  y          : {self.y}",
            f"  n          : {self.stats.n}",
            f"  deviance   : {_fmt_smart(self.stats.deviance)}",
            f"  AIC        : {_fmt_smart(self.stats.aic)}",
            "",
            "Coefficients:",
            f"  {'Term':<30}  {'Coef':>10}  {'Std.Err':>10}  {'t':>8}  {'P>|t|':>8}  {'[0.025':>10}  {'0.975]':>10}",
            f"  {'-' * 30}  {'-' * 10}  {'-' * 10}  {'-' * 8}  {'-' * 8}  {'-' * 10}  {'-' * 10}",
        ]
        for c in self.coefs:
            t_val = c.wald.value if c.wald else 0.0
            p_val = c.wald.p_value if c.wald else 1.0
            lines.append(
                f"  {c.term:<30}  {_fmt_fixed(c.est):>10}  {_fmt_fixed(c.se):>10}"
                f"  {_fmt_fixed(t_val):>8}  {_fmt_p(p_val):>8}"
                f"  {_fmt_fixed(c.lci):>10}  {_fmt_fixed(c.uci):>10}"
            )
        return "\n".join(lines)

    def __str__(self) -> str:
        P = _ui_module()
        if P:
            return P.render_rich_to_str(self, width=_resolve_width_for(self))
        return self.__plain_str__()

    def __repr__(self) -> str:
        return f"GLMFit(y={self.y!r}, family={self.family!r}, n={self.stats.n}, coefs={len(self.coefs)})"



from .logconfig import enable_debug, enable_logging, temporary_log_level  # re-export
from .random_state import RandomState, seed_from_random_state


__all__ = [
    "enable_logging",
    "enable_debug",
    "temporary_log_level",
    "RandomState",
    "seed_from_random_state",
]



# src/svy/core/describe_runtime.py
from __future__ import annotations

import datetime as dt
import logging

from typing import TYPE_CHECKING, Sequence

import polars as pl

from .describe import (
    DescribeBoolean,
    DescribeContinuous,
    DescribeDatetime,
    DescribeDiscrete,
    DescribeItem,
    DescribeNominal,
    DescribeOrdinal,
    DescribeResult,
    DescribeString,
    Freq,
    Quantile,
)
from .enumerations import MeasurementType


if TYPE_CHECKING:
    from svy.metadata import MetadataStore


log = logging.getLogger(__name__)

DEFAULT_PERCENTILES: tuple[float, ...] = (0.05, 0.25, 0.50, 0.75, 0.95)


def _infer_mtype_from_polars(dt: pl.DataType) -> MeasurementType:
    base = dt.base_type()
    if base.is_float():
        return MeasurementType.CONTINUOUS
    if base.is_integer():
        return MeasurementType.DISCRETE
    if base == pl.Boolean:
        return MeasurementType.BOOLEAN
    if base == pl.String:
        return MeasurementType.STRING
    if base == pl.Categorical or base == pl.Enum:
        return MeasurementType.NOMINAL
    if base.is_temporal():
        return MeasurementType.DATETIME
    return MeasurementType.STRING


def _resolve_columns(df: pl.DataFrame, columns: Sequence[str] | None) -> list[str]:
    if columns is None:
        return list(df.columns)
    out: list[str] = []
    names = set(df.columns)
    for c in columns:
        if c not in names:
            raise KeyError(f"Column {c!r} not found in data.")
        out.append(c)
    return out


def _series_nonnull(df: pl.DataFrame, col: str, *, drop_nulls: bool) -> pl.Series:
    s = df.get_column(col)
    return s.drop_nulls() if drop_nulls else s


def _series_missing_and_clean(s: pl.Series) -> tuple[int, pl.Series]:
    n_missing = int(s.null_count())
    clean = s.drop_nulls()
    return n_missing, clean


def _weight_series(df: pl.DataFrame, weight_col: str | None) -> pl.Series | None:
    if weight_col is None:
        return None
    if weight_col not in df.columns:
        raise KeyError(f"Weight column {weight_col!r} not found.")
    return df.get_column(weight_col).cast(pl.Float64)


def _to_float(x: object) -> float | None:
    """Safely convert a polars aggregate result to float, returning None for non-numeric."""
    if x is None:
        return None
    if isinstance(x, (int, float)):
        return float(x)
    return None


def _to_int(x: object) -> int | None:
    """Safely convert a polars aggregate result to int, returning None for non-numeric."""
    if x is None:
        return None
    if isinstance(x, (int, float)):
        return int(x)
    return None


def _to_temporal(x: object) -> dt.datetime | dt.date | dt.time | None:
    """Cast a polars min/max result to datetime/date/time or None."""
    if x is None:
        return None
    if isinstance(x, (dt.datetime, dt.date, dt.time)):
        return x
    return None


def _desc_continuous(
    df: pl.DataFrame,
    col: str,
    mtype: MeasurementType,
    *,
    weighted: bool,
    weight_col: str | None,
    drop_nulls: bool,
    percentiles: Sequence[float],
) -> DescribeContinuous:
    s_base = df.get_column(col).cast(pl.Float64)
    n_total = df.height
    n_missing, s = _series_missing_and_clean(s_base)

    mean = _to_float(s.mean()) if s.len() else None
    std = _to_float(s.std()) if s.len() else None
    var = _to_float(s.var()) if s.len() else None
    vmin = _to_float(s.min()) if s.len() else None
    vmax = _to_float(s.max()) if s.len() else None
    p25 = _to_float(s.quantile(0.25)) if s.len() else None
    p50 = _to_float(s.quantile(0.50)) if s.len() else None
    p75 = _to_float(s.quantile(0.75)) if s.len() else None
    ssum = _to_float(s.sum()) if s.len() else None

    q_items: list[Quantile] = []
    for p in percentiles:
        try:
            qv = None if weighted else (_to_float(s.quantile(p)) if s.len() else None)
        except Exception:
            qv = None
        q_items.append(Quantile(p=float(p), value=qv))

    if weighted:
        w = _weight_series(df, weight_col)
        if w is not None:
            mask = ~s_base.is_null()
            w_nonnull = w.filter(mask)
            denom = _to_float(w_nonnull.sum()) if w_nonnull.len() else 0.0
            if denom and denom > 0 and s.len():
                ssum = _to_float((s * w_nonnull).sum())
                mean = (ssum / denom) if ssum is not None else None

    return DescribeContinuous(
        name=col,
        mtype=mtype,
        n=n_total,
        n_missing=n_missing,
        weighted=weighted,
        drop_nulls=drop_nulls,
        mean=mean,
        std=std,
        var=var,
        min=vmin,
        p25=p25,
        p50=p50,
        p75=p75,
        max=vmax,
        sum=ssum,
        percentiles=tuple(q_items),
    )


def _desc_discrete(
    df: pl.DataFrame,
    col: str,
    mtype: MeasurementType,
    *,
    weighted: bool,
    weight_col: str | None,
    drop_nulls: bool,
    top_k: int,
    percentiles: Sequence[float],
) -> DescribeDiscrete:
    cont = _desc_continuous(
        df,
        col,
        mtype,
        weighted=weighted,
        weight_col=weight_col,
        drop_nulls=drop_nulls,
        percentiles=percentiles,
    )

    s_base = df.get_column(col)
    if drop_nulls:
        mask = ~s_base.is_null()
        s = s_base.filter(mask)
        w = _weight_series(df, weight_col)
        if w is not None:
            w = w.filter(mask)
    else:
        s = s_base
        w = _weight_series(df, weight_col)

    if s.len() == 0:
        levels: tuple[Freq, ...] = ()
    else:
        if weighted and w is not None:
            tbl = (
                pl.DataFrame({col: s, "_w": w})
                .group_by(col)
                .agg(pl.col("_w").sum().alias("count"))
                .sort("count", descending=True)
                .head(top_k)
            )
            denom = _to_float(tbl["count"].sum()) if tbl.height else 0.0
            levels = tuple(
                Freq(
                    level=row[col],
                    count=float(row["count"]),
                    prop=(float(row["count"]) / denom if denom and denom > 0 else 0.0),
                )
                for row in tbl.iter_rows(named=True)
            )
        else:
            tbl = (
                pl.DataFrame({col: s})
                .group_by(col)
                .len()
                .rename({"len": "count"})
                .sort("count", descending=True)
                .head(top_k)
            )
            denom = float(s.len())
            levels = tuple(
                Freq(
                    level=row[col],
                    count=float(row["count"]),
                    prop=(float(row["count"]) / denom if denom > 0 else 0.0),
                )
                for row in tbl.iter_rows(named=True)
            )

    return DescribeDiscrete(
        name=cont.name,
        mtype=cont.mtype,
        n=cont.n,
        n_missing=cont.n_missing,
        weighted=cont.weighted,
        drop_nulls=cont.drop_nulls,
        mean=cont.mean,
        std=cont.std,
        var=cont.var,
        min=cont.min,
        p25=cont.p25,
        p50=cont.p50,
        p75=cont.p75,
        max=cont.max,
        sum=cont.sum,
        levels=levels,
    )


def _build_freqs(
    s: pl.Series,
    *,
    weighted: bool,
    w: pl.Series | None,
    top_k: int,
) -> tuple[Freq, ...]:
    if s.len() == 0:
        return ()
    if weighted and w is not None:
        tbl = (
            pl.DataFrame({"__k": s, "__w": w})
            .group_by("__k")
            .agg(pl.col("__w").sum().alias("count"))
            .sort("count", descending=True)
            .head(top_k)
        )
        denom = _to_float(tbl["count"].sum()) if tbl.height else 0.0
        return tuple(
            Freq(
                level=row["__k"],
                count=float(row["count"]),
                prop=(float(row["count"]) / denom if denom and denom > 0 else 0.0),
            )
            for row in tbl.iter_rows(named=True)
        )
    else:
        tbl = (
            pl.DataFrame({"__k": s})
            .group_by("__k")
            .len()
            .rename({"len": "count"})
            .sort("count", descending=True)
            .head(top_k)
        )
        denom = float(s.len())
        return tuple(
            Freq(
                level=row["__k"],
                count=float(row["count"]),
                prop=(float(row["count"]) / denom if denom > 0 else 0.0),
            )
            for row in tbl.iter_rows(named=True)
        )


def _desc_nominal_or_ordinal(
    df: pl.DataFrame,
    col: str,
    mtype: MeasurementType,
    *,
    weighted: bool,
    weight_col: str | None,
    drop_nulls: bool,
    top_k: int,
    ordinal: bool,
) -> DescribeNominal | DescribeOrdinal:
    s_base = df.get_column(col)
    if drop_nulls:
        mask = ~s_base.is_null()
        s = s_base.filter(mask)
        w = _weight_series(df, weight_col)
        if w is not None:
            w = w.filter(mask)
    else:
        s = s_base
        w = _weight_series(df, weight_col)

    n_total = df.height
    n_missing = int(s_base.null_count())

    levels = _build_freqs(s, weighted=weighted, w=w, top_k=top_k)
    n_levels = int(len(levels))
    mode = levels[0].level if levels else None

    if ordinal:
        return DescribeOrdinal(
            name=col,
            mtype=MeasurementType.ORDINAL,
            n=n_total,
            n_missing=n_missing,
            weighted=weighted,
            drop_nulls=drop_nulls,
            levels=levels,
            n_levels=n_levels,
            mode=mode,
            truncated=False,
        )
    return DescribeNominal(
        name=col,
        mtype=MeasurementType.NOMINAL,
        n=n_total,
        n_missing=n_missing,
        weighted=weighted,
        drop_nulls=drop_nulls,
        levels=levels,
        n_levels=n_levels,
        mode=mode,
        truncated=False,
    )


def _desc_boolean(
    df: pl.DataFrame,
    col: str,
    mtype: MeasurementType,
    *,
    weighted: bool,
    weight_col: str | None,
    drop_nulls: bool,
) -> DescribeBoolean:
    s_base = df.get_column(col).cast(pl.Boolean)
    if drop_nulls:
        mask = ~s_base.is_null()
        s = s_base.filter(mask)
        w = _weight_series(df, weight_col)
        if w is not None:
            w = w.filter(mask)
    else:
        s = s_base
        w = _weight_series(df, weight_col)

    n_total = df.height
    n_missing = int(s_base.null_count())

    if s.len() == 0:
        f_false = None
        f_true = None
    else:
        if weighted and w is not None:
            c_true = float(
                pl.DataFrame({"b": s.cast(pl.Int8), "w": w})
                .with_columns((pl.col("b") * pl.col("w")).alias("wt"))
                .select(pl.col("wt").sum())
                .item()
            )
            c_false = float(w.sum()) - c_true
            denom = c_true + c_false
        else:
            c_true = float(s.sum())
            c_false = float(s.len() - s.sum())
            denom = float(s.len())

        p_true = (c_true / denom) if denom > 0 else 0.0
        p_false = (c_false / denom) if denom > 0 else 0.0
        f_true = Freq(level=True, count=c_true, prop=p_true)
        f_false = Freq(level=False, count=c_false, prop=p_false)

    return DescribeBoolean(
        name=col,
        mtype=MeasurementType.BOOLEAN,
        n=n_total,
        n_missing=n_missing,
        weighted=weighted,
        drop_nulls=drop_nulls,
        false=f_false,
        true=f_true,
    )


def _desc_datetime(
    df: pl.DataFrame,
    col: str,
    mtype: MeasurementType,
    *,
    weighted: bool,
    drop_nulls: bool,
) -> DescribeDatetime:
    s_base = df.get_column(col)
    n_total = df.height
    n_missing = int(s_base.null_count())
    s = s_base.drop_nulls() if drop_nulls else s_base

    try:
        vmin = _to_temporal(s.min())
        vmax = _to_temporal(s.max())
    except Exception:
        vmin = None
        vmax = None

    return DescribeDatetime(
        name=col,
        mtype=MeasurementType.DATETIME,
        n=n_total,
        n_missing=n_missing,
        weighted=weighted,
        drop_nulls=drop_nulls,
        min=vmin,
        max=vmax,
        tz=None,
    )


def _desc_string(
    df: pl.DataFrame,
    col: str,
    mtype: MeasurementType,
    *,
    weighted: bool,
    weight_col: str | None,
    drop_nulls: bool,
    top_k: int,
) -> DescribeString:
    s_base = df.get_column(col).cast(pl.Utf8)
    if drop_nulls:
        mask = ~s_base.is_null()
        s = s_base.filter(mask)
        w = _weight_series(df, weight_col)
        if w is not None:
            w = w.filter(mask)
    else:
        s = s_base
        w = _weight_series(df, weight_col)

    n_total = df.height
    n_missing = int(s_base.null_count())

    top = _build_freqs(s, weighted=weighted, w=w, top_k=top_k)
    try:
        n_unique = int(s.n_unique())
    except Exception:
        n_unique = None

    s_len_base = s.drop_nulls()
    if s_len_base.len():
        ls = s_len_base.str.len_chars()
        len_min = _to_int(ls.min())
        len_max = _to_int(ls.max())
        len_mean = _to_float(ls.mean())
        len_p50 = _to_float(ls.quantile(0.5))
    else:
        len_min = len_max = None
        len_mean = len_p50 = None

    return DescribeString(
        name=col,
        mtype=MeasurementType.STRING,
        n=n_total,
        n_missing=n_missing,
        weighted=weighted,
        drop_nulls=drop_nulls,
        n_unique=n_unique,
        top=top,
        truncated=False,
        len_min=len_min,
        len_p50=len_p50,
        len_mean=len_mean,
        len_max=len_max,
    )


def run_describe(
    *,
    df: pl.DataFrame,
    metadata: "MetadataStore | None" = None,
    columns: Sequence[str] | None = None,
    weighted: bool = False,
    weight_col: str | None = None,
    drop_nulls: bool = True,
    top_k: int = 10,
    percentiles: Sequence[float] = DEFAULT_PERCENTILES,
) -> DescribeResult:
    cols = _resolve_columns(df, columns)

    items: list[DescribeItem] = []
    for col in cols:
        mtype: MeasurementType
        if metadata is not None:
            meta = metadata.get(col)
            if meta is not None:
                mtype = meta.mtype
            else:
                mtype = _infer_mtype_from_polars(df.schema[col])
        else:
            mtype = _infer_mtype_from_polars(df.schema[col])

        if mtype is MeasurementType.CONTINUOUS:
            items.append(
                _desc_continuous(
                    df,
                    col,
                    mtype,
                    weighted=weighted,
                    weight_col=weight_col,
                    drop_nulls=drop_nulls,
                    percentiles=percentiles,
                )
            )
        elif mtype is MeasurementType.DISCRETE:
            items.append(
                _desc_discrete(
                    df,
                    col,
                    mtype,
                    weighted=weighted,
                    weight_col=weight_col,
                    drop_nulls=drop_nulls,
                    top_k=top_k,
                    percentiles=percentiles,
                )
            )
        elif mtype is MeasurementType.NOMINAL:
            items.append(
                _desc_nominal_or_ordinal(
                    df,
                    col,
                    mtype,
                    weighted=weighted,
                    weight_col=weight_col,
                    drop_nulls=drop_nulls,
                    top_k=top_k,
                    ordinal=False,
                )
            )
        elif mtype is MeasurementType.ORDINAL:
            items.append(
                _desc_nominal_or_ordinal(
                    df,
                    col,
                    mtype,
                    weighted=weighted,
                    weight_col=weight_col,
                    drop_nulls=drop_nulls,
                    top_k=top_k,
                    ordinal=True,
                )
            )
        elif mtype is MeasurementType.BOOLEAN:
            items.append(
                _desc_boolean(
                    df, col, mtype, weighted=weighted, weight_col=weight_col, drop_nulls=drop_nulls
                )
            )
        elif mtype is MeasurementType.DATETIME:
            items.append(_desc_datetime(df, col, mtype, weighted=weighted, drop_nulls=drop_nulls))
        else:
            items.append(
                _desc_string(
                    df,
                    col,
                    mtype,
                    weighted=weighted,
                    weight_col=weight_col,
                    drop_nulls=drop_nulls,
                    top_k=top_k,
                )
            )

    return DescribeResult(
        items=tuple(items),
        weighted=weighted,
        weight_col=weight_col,
        drop_nulls=drop_nulls,
        top_k=int(top_k),
        percentiles=tuple(float(p) for p in percentiles),
        generated_at=dt.datetime.now(dt.timezone.utc),
        notes=None,
    )



# src/svy/evaluation/compare.py
from __future__ import annotations

from typing import Any, Sequence

import numpy as np
import polars as pl

from rich.table import Table as RTable
from rich.text import Text

from svy.regression.glm import GLM
from svy.regression.glm_pred import GLMPred  # type: ignore[unresolved-import]
from svy.utils.formats import _fmt_fixed


def compare(
    objects: Sequence[Any],
    names: Sequence[str] | None = None,
    **kwargs,
) -> Any:
    """
    Universal comparison function.
    Dispatches to specific comparison logic based on input types.

    Parameters
    ----------
    objects : Sequence[Any]
        A list of objects to compare (GLMs, Predictions, etc).
    names : Sequence[str] | None
        Optional labels for the columns/rows.
    **kwargs :
        Additional arguments passed to the specific comparison function
        (e.g. metrics=["aic", "bic"] for models, metrics=["mse"] for preds).
    """
    if not objects:
        raise ValueError("No objects provided to compare.")

    # 1. Identify Type of the first element
    first = objects[0]

    # 2. Dispatch: GLM Models
    if isinstance(first, GLM):
        # Validation: Ensure all are GLMs
        if not all(isinstance(x, GLM) for x in objects):
            raise TypeError("Cannot compare mixed types. All objects must be GLM instances.")
        return compare_models(objects, names=names, **kwargs)

    # 3. Dispatch: Predictions (GLMPred class)
    elif isinstance(first, GLMPred):
        if not all(isinstance(x, GLMPred) for x in objects):
            raise TypeError("All objects must be GLMPred instances.")
        return compare_predictions(objects, names=names, **kwargs)

    # 4. Dispatch: Raw Arrays (e.g. comparing y_true vs multiple y_pred)
    elif isinstance(first, (np.ndarray, pl.Series, list)):
        return compare_arrays(objects, names=names, **kwargs)

    else:
        obj_type = type(first).__name__
        raise NotImplementedError(f"No comparison logic implemented for type '{obj_type}'")


# =============================================================================
# Model Comparison Implementation
# =============================================================================


class ModelComparison:
    """Container for model comparison results with publication-quality printing."""

    def __init__(self, coefs: pl.DataFrame, stats: pl.DataFrame, names: list[str]):
        self.coefs = coefs
        self.stats = stats
        self.names = names

    def __repr__(self) -> str:
        return f"<ModelComparison: {len(self.names)} models>"

    def __rich_console__(self, console, options):
        # Create a "Stata/Academic-style" table
        tbl = RTable(show_header=True, header_style="bold magenta", box=None, padding=(0, 2))

        # Columns: Term, Model 1, Model 2...
        tbl.add_column("Term", style="bold")
        for name in self.names:
            tbl.add_column(name, justify="center")

        # 1. Coefficients
        # We iterate through Polars rows
        for row in self.coefs.iter_rows(named=True):
            cells = [row["term"]]

            for name in self.names:
                est = row[f"{name}_est"]
                se = row[f"{name}_se"]

                if est is None:
                    cells.append("-")
                else:
                    # Format:
                    #  0.123
                    #  (0.05)
                    est_str = _fmt_fixed(est)
                    se_str = f"({_fmt_fixed(se)})"

                    # Combine into one cell with newline, dim the SE
                    cells.append(Text.assemble(est_str, "\n", (se_str, "dim")))

            tbl.add_row(*cells)

        # 2. Divider (Visual separation between coefs and stats)
        tbl.add_section()

        # 3. Statistics
        for row in self.stats.iter_rows(named=True):
            stat_label = row["stat"].replace("_", " ").title()
            cells = [Text(stat_label, style="italic")]

            for name in self.names:
                val = row[name]
                if val is None:
                    cells.append(Text("-"))
                elif isinstance(val, int) or (isinstance(val, float) and val.is_integer()):
                    cells.append(Text(str(int(val))))
                else:
                    cells.append(Text(_fmt_fixed(val)))
            tbl.add_row(*cells)

        yield tbl


def compare_models(
    models: Sequence[GLM],
    names: Sequence[str] | None,
    metrics: Sequence[str] = ("n", "aic", "bic", "r_squared_adj"),
) -> ModelComparison:
    """
    Compare multiple fitted GLM models side-by-side.
    """
    # 1. Resolve Names
    if names is None:
        names = [f"Model {i + 1}" for i in range(len(models))]

    if len(names) != len(models):
        raise ValueError(
            f"Length of 'names' ({len(names)}) must match length of 'models' ({len(models)})."
        )

    # 2. Extract Coefficients
    # We want to align coefficients by term name across all models
    term_map: dict[str, dict[str, tuple[float, float]]] = {}
    all_terms_order = []

    for name, model in zip(names, models):
        if model.fitted is None:
            continue

        for c in model.coefs:
            if c.term not in term_map:
                term_map[c.term] = {}
                all_terms_order.append(c.term)

            term_map[c.term][name] = (c.est, c.se)

    # 3. Build Polars DataFrame for Coefs
    # Structure: [term, m1_est, m1_se, m2_est, m2_se, ...]
    data = {"term": all_terms_order}
    for name in names:
        ests = []
        ses = []
        for term in all_terms_order:
            val = term_map[term].get(name)
            if val:
                ests.append(val[0])
                ses.append(val[1])
            else:
                ests.append(None)
                ses.append(None)

        data[f"{name}_est"] = ests
        data[f"{name}_se"] = ses

    df_coefs = pl.DataFrame(data)

    # 4. Extract Model-Level Stats
    stats_data = []
    for m in metrics:
        row: dict[str, Any] = {"stat": m}
        for name, model in zip(names, models):
            if model.fitted is None:
                row[name] = None
                continue

            # Use getattr to fetch stats dynamically from GLMStats struct
            # GLM.stats property returns GLMStats object
            val = getattr(model.stats, m, None)
            row[name] = val
        stats_data.append(row)

    df_stats = pl.DataFrame(stats_data)

    return ModelComparison(df_coefs, df_stats, list(names))


# =============================================================================
# Other Implementations (Stubs)
# =============================================================================


def compare_predictions(
    preds: Sequence[GLMPred], names: Sequence[str] | None, metrics: Sequence[str] = ("rmse", "mae")
):
    # TODO: Implement prediction comparison table
    print(f"Comparing {len(preds)} Predictions (Not implemented yet)...")


def compare_arrays(arrays: Sequence[Any], names: Sequence[str] | None, **kwargs):
    # TODO: Implement raw array comparison
    pass



# src/svy/questionnaire/base.py
from __future__ import annotations

import logging

from enum import Enum
from typing import Any, Mapping, Self, cast

import msgspec

from svy.core.types import Category
from svy.metadata.labels import Label, LabellingCatalog, MissingKind


log = logging.getLogger(__name__)

# =============================
# Question types & conditions
# =============================


class QuestionType(Enum):
    SINGLE = "single"  # single-select categorical
    MULTI = "multi"  # multi-select categorical
    BOOLEAN = "boolean"  # yes/no (can still use a scheme)
    NUMERIC = "numeric"  # integers/floats
    TEXT = "text"  # free text
    DATE = "date"  # ISO string yyyy-mm-dd (keep simple)


class Op(Enum):
    EQ = "=="
    NE = "!="
    IN = "in"
    NOT_IN = "not_in"
    GT = ">"
    GE = ">="
    LT = "<"
    LE = "<="
    IS_MISSING = "is_missing"
    NOT_MISSING = "not_missing"


class EnableWhen(msgspec.Struct, frozen=True):
    """Minimal, serializable condition for routing/enablement."""

    question_id: str
    op: Op
    value: Category | list[Category] | None = None


# =============================
# Choice source
# =============================


class Choices(msgspec.Struct, frozen=True):
    """
    Choices can come from:
      - inline mapping (mapping=...)
      - a catalog concept (concept=..., locale=...)
    Provide exactly one of (mapping, concept).
    """

    mapping: dict[Category, str] | None = None
    concept: str | None = None
    locale: str | None = None
    title: str | None = None
    randomize: bool = False
    allow_other: bool = False  # UI hint; 'other' not encoded here

    def resolve_mapping(self, catalog: LabellingCatalog | None) -> dict[Category, str]:
        if self.mapping is not None:
            # Return a shallow copy to preserve immutability expectations
            return dict(self.mapping)
        if self.concept is None:
            return {}
        if catalog is None:
            raise ValueError("Choices reference a concept but no catalog was provided.")
        lbl: Label = catalog.to_label_by_concept(
            var_label=self.title or self.concept, concept=self.concept, locale=self.locale
        )
        # Be explicit for the type checker: categories may be Optional/Mapping-like
        cats: Mapping[Category, str] | None = getattr(lbl, "categories", None)
        return dict(cats) if cats is not None else {}


# =============================
# Questions & Questionnaire
# =============================


class Question(msgspec.Struct, frozen=True):
    """
    A single question. 'name' is the variable name/column id (stable).
    """

    name: str
    text: str
    qtype: QuestionType
    choices: Choices | None = None  # for categorical types
    required: bool = False

    # Numeric/text constraints (optional)
    min_value: float | None = None
    max_value: float | None = None
    min_length: int | None = None
    max_length: int | None = None

    # Skip logic (all conditions must be satisfied for the question to be enabled).
    enable_when_all: list[EnableWhen] | None = None

    # Which missing kinds are permitted to be recorded for this question (optional)
    allow_missing_kinds: set[MissingKind] | None = None

    def validate(self) -> None:
        # type-consistency checks
        if self.qtype in (QuestionType.SINGLE, QuestionType.MULTI, QuestionType.BOOLEAN):
            # choices optional for BOOLEAN; recommended for SINGLE/MULTI
            if self.qtype != QuestionType.BOOLEAN and self.choices is None:
                raise ValueError(f"Question {self.name!r}: categorical type requires choices.")
        else:
            if self.choices is not None:
                raise ValueError(
                    f"Question {self.name!r}: choices not allowed for {self.qtype.value}."
                )

        if self.qtype == QuestionType.NUMERIC:
            if self.min_value is not None and self.max_value is not None:
                if self.min_value > self.max_value:
                    raise ValueError(f"Question {self.name!r}: min_value > max_value.")
        if self.qtype == QuestionType.TEXT:
            if self.min_length is not None and self.max_length is not None:
                if self.min_length > self.max_length:
                    raise ValueError(f"Question {self.name!r}: min_length > max_length.")

    def is_enabled(self, answers: Mapping[str, Any]) -> bool:
        """Check skip logic against already-collected answers."""
        conds = self.enable_when_all or []
        for c in conds:
            val = answers.get(c.question_id, None)

            if c.op == Op.IS_MISSING:
                if val is not None:
                    return False
                continue
            if c.op == Op.NOT_MISSING:
                if val is None:
                    return False
                continue

            if c.op in (Op.EQ, Op.NE, Op.GT, Op.GE, Op.LT, Op.LE):
                # numeric/text equality/ordering; simple compare
                if c.op == Op.EQ and not (val == c.value):
                    return False
                if c.op == Op.NE and not (val != c.value):
                    return False
                if c.op == Op.GT and not (
                    val is not None and c.value is not None and val > c.value
                ):
                    return False
                if c.op == Op.GE and not (
                    val is not None and c.value is not None and val >= c.value
                ):
                    return False
                if c.op == Op.LT and not (
                    val is not None and c.value is not None and val < c.value
                ):
                    return False
                if c.op == Op.LE and not (
                    val is not None and c.value is not None and val <= c.value
                ):
                    return False
                continue

            if c.op == Op.IN:
                if not isinstance(c.value, list) or val not in c.value:
                    return False
                continue
            if c.op == Op.NOT_IN:
                if isinstance(c.value, list) and val in c.value:
                    return False
                continue

            raise ValueError(f"Unsupported op {c.op} in condition for {self.name!r}.")
        return True

    def resolved_choices(self, catalog: LabellingCatalog | None) -> dict[Category, str]:
        # The empty dict literal needs a cast for precise key type (Category)
        return (
            self.choices.resolve_mapping(catalog)
            if self.choices
            else cast(dict[Category, str], {})
        )

    def clone(self, **overrides) -> Self:
        from msgspec.structs import replace

        return replace(self, **overrides)


class Questionnaire(msgspec.Struct):
    """
    Minimal questionnaire container.
    - Store metadata
    - Hold questions (ordered)
    - Provide resolution/validation helpers
    """

    title: str
    locale: str | None = None
    version: str | None = None
    questions: list[Question] = msgspec.field(default_factory=list)

    # ---------- management ----------
    def add(self, *qs: Question) -> Self:
        self.questions.extend(qs)
        return self

    def add_question(self, **kwargs: Any) -> Self:
        """
        Build-and-append from kwargs:
        Questionnaire().add_question(
            name="consent",
            text="Do you agree to participate?",
            qtype=QuestionType.SINGLE,
            choices=Choices(concept="yes_no"),
            required=True,
        )
        """
        q = Question(**kwargs)
        q.validate()
        self.questions.append(q)
        return self

    def add_questions(self, *items: dict[str, Any]) -> Self:
        """Batch add from kwargs dicts."""
        for kw in items:
            self.add_question(**kw)
        return self

    def remove(self, name: str) -> Self:
        """Remove a question by name."""
        original_len = len(self.questions)
        self.questions = [q for q in self.questions if q.name != name]
        if len(self.questions) == original_len:
            raise KeyError(f"Question not found: {name!r}")
        return self

    def update(self, name: str, **kwargs: Any) -> Self:
        """Update a question's fields by name."""
        if "name" in kwargs:
            raise ValueError("Cannot change question name via update().")
        for i, q in enumerate(self.questions):
            if q.name == name:
                self.questions[i] = q.clone(**kwargs)
                return self
        raise KeyError(f"Question not found: {name!r}")

    def find(self, name: str) -> Question:
        for q in self.questions:
            if q.name == name:
                return q
        raise KeyError(name)

    # ---------- validation ----------
    def validate(self) -> None:
        seen = set()
        for q in self.questions:
            if q.name in seen:
                raise ValueError(f"Duplicate question name: {q.name!r}")
            seen.add(q.name)
            q.validate()

    # ---------- rendering helpers ----------
    def choice_map(self, name: str, catalog: LabellingCatalog | None) -> dict[Category, str]:
        return self.find(name).resolved_choices(catalog)

    def to_codebook(self, catalog: LabellingCatalog | None) -> list[dict[str, Any]]:
        """
        Produce a simple codebook for UI/exports.
        """
        out: list[dict[str, Any]] = []
        for q in self.questions:
            row: dict[str, Any] = {
                "name": q.name,
                "text": q.text,
                "type": q.qtype.value,
            }
            cm = q.resolved_choices(catalog)
            if cm:
                row["choices"] = cm
            out.append(row)
        return out

    # ---------- simple answering ----------
    def apply_skip_logic(self, answers: Mapping[str, Any]) -> set[str]:
        """
        Return the set of question names that are enabled given current answers.
        """
        enabled: set[str] = set()
        for q in self.questions:
            if q.is_enabled(answers):
                enabled.add(q.name)
        return enabled

    def validate_answer(
        self, name: str, value: Any, catalog: LabellingCatalog | None = None
    ) -> None:
        q = self.find(name)
        if value is None:
            if q.required:
                raise ValueError(f"{name} is required.")
            return

        if q.qtype == QuestionType.SINGLE or q.qtype == QuestionType.BOOLEAN:
            cm = q.resolved_choices(catalog)
            if value not in cm:
                raise ValueError(f"{name}: invalid code {value!r}. Allowed: {list(cm.keys())}")
        elif q.qtype == QuestionType.MULTI:
            cm = q.resolved_choices(catalog)
            if not isinstance(value, list):
                raise ValueError(f"{name}: expected list for MULTI.")
            bad = [v for v in value if v not in cm]
            if bad:
                raise ValueError(f"{name}: invalid codes {bad!r}. Allowed: {list(cm.keys())}")
        elif q.qtype == QuestionType.NUMERIC:
            if not isinstance(value, (int, float)):
                raise ValueError(f"{name}: expected numeric.")
            if q.min_value is not None and value < q.min_value:
                raise ValueError(f"{name}: value < min_value.")
            if q.max_value is not None and value > q.max_value:
                raise ValueError(f"{name}: value > max_value.")
        elif q.qtype == QuestionType.TEXT:
            if not isinstance(value, str):
                raise ValueError(f"{name}: expected text.")
            n = len(value)
            if q.min_length is not None and n < q.min_length:
                raise ValueError(f"{name}: text shorter than min_length.")
            if q.max_length is not None and n > q.max_length:
                raise ValueError(f"{name}: text longer than max_length.")
        elif q.qtype == QuestionType.DATE:
            if not isinstance(value, str):
                raise ValueError(f"{name}: expected ISO date string.")
        else:
            raise ValueError(f"Unsupported question type: {q.qtype}.")

    def to_row(self, answers: Mapping[str, Any]) -> dict[str, Any]:
        """
        Convert validated answers to a row (dict). For MULTI, store as a list (up to caller to expand).
        This function assumes `validate_answer` has been called per item.
        """
        row: dict[str, Any] = {}
        for q in self.questions:
            row[q.name] = answers.get(q.name, None)
        return row



# src/svy/weighting/base.py
from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Any, Literal, Mapping, Sequence, cast

import numpy as np
import polars as pl


try:
    from svy_rs._internal import (  # type: ignore[import-untyped]
        adjust_nr as rust_adjust_nr,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        calibrate as rust_calibrate,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        calibrate_by_domain as rust_calibrate_by_domain,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        calibrate_parallel as rust_calibrate_parallel,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        create_bootstrap_wgts as rust_create_bootstrap_wgts,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        create_brr_wgts as rust_create_brr_wgts,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        create_jk_wgts as rust_create_jk_wgts,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        create_sdr_wgts as rust_create_sdr_wgts,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        normalize as rust_normalize,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        poststratify as rust_poststratify,
    )
    from svy_rs._internal import (  # type: ignore[import-untyped]
        rake as rust_rake,
    )
except ImportError:  # pragma: no cover
    rust_adjust_nr = rust_calibrate = rust_calibrate_by_domain = None
    rust_calibrate_parallel = rust_create_bootstrap_wgts = None
    rust_create_brr_wgts = rust_create_jk_wgts = rust_create_sdr_wgts = None
    rust_normalize = rust_poststratify = rust_rake = None

from svy.core.constants import (
    SVY_ADJ_PREFIX,
    SVY_CALIB_PREFIX,
    SVY_NORM_PREFIX,
    SVY_PS_PREFIX,
    SVY_RAKED_PREFIX,
)
from svy.core.design import RepWeights
from svy.core.enumerations import EstimationMethod
from svy.core.terms import Cat, Cross, Feature
from svy.core.types import (
    Category,
    ControlsType,
    DomainScalarMap,
    NDArray,
    Number,
)
from svy.engine.weighting.adj_calibration import _calibrate
from svy.engine.weighting.adj_nonresponse import _adjust_nr
from svy.engine.weighting.adj_normalization import _normalize, _normalize_controls_like
from svy.engine.weighting.adj_poststratification import _poststratify
from svy.engine.weighting.adj_raking import _calculate_controls_from_factors, _rake
from svy.errors import DimensionError, MethodError
from svy.utils.checks import drop_missing
from svy.utils.random_state import RandomState, resolve_random_state


if TYPE_CHECKING:
    from svy.core.sample import Sample

log = logging.getLogger(__name__)


# ---------------------------
# Internal Helpers
# ---------------------------


def _to_int_array(
    data: pl.DataFrame,
    col_spec: str | tuple[str, ...] | None,
) -> np.ndarray | None:
    """
    Convert a single or multi-column design field to a 1-D int64 array
    suitable for passing to Rust functions.

    Parameters
    ----------
    data     : source DataFrame
    col_spec : str, tuple[str, ...], or None
               - None          → returns None
               - str           → encodes that single column
               - tuple[str,…]  → concatenates columns into composite labels,
                                  then integer-encodes

    Returns
    -------
    np.ndarray[int64] | None
    """
    if col_spec is None:
        return None
    cols = [col_spec] if isinstance(col_spec, str) else list(col_spec)
    if len(cols) == 1:
        vec = data[cols[0]].to_numpy()
    else:
        # Concatenate multi-column keys into a single composite string label
        vec = np.array(
            ["__".join(str(v) for v in row) for row in data.select(cols).iter_rows()],
            dtype=object,
        )
    if vec.dtype.kind in ("U", "S", "O"):
        _, codes = np.unique(vec, return_inverse=True)
        return codes.astype(np.int64)
    return vec.astype(np.int64)


def _to_float_array(
    data: pl.DataFrame,
    col: str | None,
    n: int,
) -> np.ndarray:
    """
    Return a float64 weight array from a column, or ones if col is None.
    Always returns float64 regardless of source dtype.
    """
    if col and col in data.columns:
        return data[col].to_numpy().astype(np.float64, copy=False)
    return np.ones(n, dtype=np.float64)


def _canonical_psu_table(df: pl.DataFrame, by_cols: list[str]) -> pl.DataFrame:
    return df.select(by_cols).unique(maintain_order=True).sort(by_cols)


def _name_rep_cols(prefix: str, n_reps: int) -> list[str]:
    return [f"{prefix}{i}" for i in range(1, n_reps + 1)]


def _ensure_columns(df: pl.DataFrame, cols: list[str], *, where: str) -> None:
    missing = [c for c in cols if c not in df.columns]
    if missing:
        raise MethodError.invalid_choice(
            where=where,
            param="columns",
            got=missing,
            allowed=df.columns,
            hint="Check spelling or pass existing column names.",
        )


def _unique_name(base: str, existing: set[str]) -> str:
    if base not in existing:
        return base
    k = 1
    while f"{base}_{k}" in existing:
        k += 1
    return f"{base}_{k}"


def _extract_cols_from_term(term: Feature) -> list[str]:
    if isinstance(term, str):
        return [term]
    if isinstance(term, Cat):
        return [term.name]
    if isinstance(term, Cross):
        return _extract_cols_from_term(term.left) + _extract_cols_from_term(term.right)
    raise TypeError(f"Unsupported term type for grouping: {type(term)}")


def _normalize_by_term(
    df: pl.DataFrame, by: Feature | None, *, param_name: str = "by", where: str
) -> tuple[list[str] | None, NDArray | None]:
    if by is None:
        return None, None

    if isinstance(by, (list, tuple)):
        raise TypeError(
            f"`{param_name}` must be a single Feature (str, Cat, Cross). "
            "Lists are not allowed; use Cross() for interactions."
        )

    adj_cols = _extract_cols_from_term(by)

    missing = [c for c in adj_cols if c not in df.columns]
    if missing:
        raise TypeError(f"`{param_name}` term refers to columns not present in data: {missing!r}")

    if len(adj_cols) == 1:
        adj_class_arr = df.get_column(adj_cols[0]).to_numpy()
    else:
        cols_np = [df.get_column(c).to_numpy() for c in adj_cols]
        adj_class_arr = np.fromiter(
            (tuple(row) for row in zip(*cols_np)),
            dtype=object,
            count=df.height,
        )

    return adj_cols, adj_class_arr


def _num_sort_key_token(tok: str) -> tuple[int, float | str]:
    try:
        return (0, float(tok))
    except Exception:
        return (1, tok)


def _num_sort_key_label(lbl):
    if isinstance(lbl, tuple):
        return tuple(_num_sort_key_token(t) for t in lbl)
    return _num_sort_key_token(lbl)


# ---------------------------
# Term Expansion & Target Matching (for Calibration)
# ---------------------------


def _expand_term(
    term: Feature, df: pl.DataFrame, where: str
) -> tuple[list[pl.Expr], list[Category]]:
    if isinstance(term, str):
        if term not in df.columns:
            raise KeyError(f"{where}: Continuous term '{term}' not found in data.")
        return [pl.col(term).cast(pl.Float64).fill_null(0.0)], [term]

    if isinstance(term, Cat):
        col_name = term.name
        if col_name not in df.columns:
            raise KeyError(f"{where}: Cat variable '{col_name}' not found.")

        levels = df.get_column(col_name).unique().sort().to_list()

        if term.ref is not None:
            if term.ref not in levels:
                raise ValueError(
                    f"{where}: Reference level '{term.ref}' not found in '{col_name}'."
                )
            levels = [lbl for lbl in levels if lbl != term.ref]

        exprs = []
        labels = []
        for i, lvl in enumerate(levels):
            if lvl is None:
                base_expr = pl.col(col_name).is_null()
            else:
                base_expr = pl.col(col_name) == lvl

            expr = base_expr.cast(pl.Float64).fill_null(0.0).alias(f"_tmp_cat_{col_name}_{i}")
            exprs.append(expr)
            labels.append(lvl)

        return exprs, labels

    if isinstance(term, Cross):
        left_exprs, left_labs = _expand_term(term.left, df, where)
        right_exprs, right_labs = _expand_term(term.right, df, where)

        out_exprs = []
        out_labs = []

        count = 0
        for le, ll in zip(left_exprs, left_labs):
            for re, rl in zip(right_exprs, right_labs):
                out_exprs.append((le * re).alias(f"_tmp_cross_{count}"))
                count += 1

                def _to_tuple(x):
                    return x if isinstance(x, tuple) else (x,)

                new_lab = _to_tuple(ll) + _to_tuple(rl)
                out_labs.append(new_lab)

        return out_exprs, out_labs

    raise TypeError(f"Unsupported term type: {type(term)}")


def _match_term_targets(
    labels: list[Category],
    target_spec: Number | dict[Category, Number] | Sequence[Number],
    term_name: str,
) -> list[float]:
    if isinstance(target_spec, (int, float, np.integer, np.floating)):
        if len(labels) != 1:
            raise ValueError(
                f"Scalar target provided for term '{term_name}' which expanded "
                f"to {len(labels)} columns. Use a dict mapping labels to values."
            )
        return [float(target_spec)]

    if isinstance(target_spec, dict):
        targets = []
        missing = [lbl for lbl in labels if lbl not in target_spec]
        if missing:
            raise ValueError(f"Missing targets for term '{term_name}': {missing}")

        for lbl in labels:
            val = target_spec[lbl]  # type: ignore[index]
            if not isinstance(val, (int, float, np.integer, np.floating)):
                raise TypeError(f"Target for '{term_name}':{lbl} must be numeric.")
            targets.append(float(val))
        return targets

    raise TypeError(
        f"Unsupported target specification type for '{term_name}': {type(target_spec)}"
    )


# ---------------------------
# Weighting Facade
# ---------------------------


class Weighting:
    def __init__(self, sample: Any) -> None:
        self._sample = sample

    def create_variance_strata(
        self,
        *,
        method: Literal["brr", "jk2"],
        order_by: str | Sequence[str] | None = None,
        shuffle: bool = False,
        into: str = "svy_var_stratum",
        rstate: int | None = None,
    ) -> Any:
        """
        Create variance estimation strata for BRR or JK2.

        Converts a design with multiple PSUs per stratum into one suitable
        for BRR (exactly 2 PSUs per variance stratum) or JK2 (2-3 PSUs).

        Args:
            method: Variance estimation method:
                - "brr": Creates pairs (2 PSUs each). Requires even PSU counts.
                - "jk2": Creates pairs, allows triplet for odd counts (2-3 PSUs each).
            order_by: Column(s) to sort PSUs by before pairing.
            shuffle: If True, randomly shuffle PSUs before pairing.
            into: Name for the new variance stratum column.
            rstate: Random seed for shuffle=True.

        Returns:
            Sample with new variance stratum column added and design updated.
        """
        where = "Sample.weighting.create_variance_strata"
        df = self._sample._data
        design = self._sample._design

        if design.psu is None:
            raise MethodError.not_applicable(
                where=where,
                method="create_variance_strata",
                reason="Design must have PSU defined.",
            )

        _method = method.lower()
        if _method not in ("brr", "jk2"):
            raise MethodError.invalid_choice(
                where=where,
                param="method",
                got=_method,
                allowed=["brr", "jk2"],
            )

        psu_col = design.psu
        orig_stratum_col = design.stratum

        if order_by is None:
            order_cols = []
        elif isinstance(order_by, str):
            order_cols = [order_by]
        else:
            order_cols = list(order_by)

        missing_cols = [col for col in order_cols if col not in df.columns]
        if missing_cols:
            raise MethodError.invalid_choice(
                where=where,
                param="order_by",
                got=missing_cols,
                allowed=df.columns,
                hint="Check column names.",
            )

        select_cols = [psu_col]
        if orig_stratum_col:
            select_cols.append(orig_stratum_col)
        select_cols.extend([c for c in order_cols if c not in select_cols])

        psu_df = df.select(select_cols).unique(maintain_order=True)

        sort_cols = []
        if orig_stratum_col:
            sort_cols.append(orig_stratum_col)
        sort_cols.extend(order_cols)
        if psu_col not in sort_cols:
            sort_cols.append(psu_col)

        psu_df = psu_df.sort(sort_cols)

        psu_list = psu_df[psu_col].to_list()
        n_psus = len(psu_list)

        if orig_stratum_col:
            orig_strata = psu_df[orig_stratum_col].to_numpy()
        else:
            orig_strata = np.zeros(n_psus, dtype=np.int64)

        unique_orig, counts = np.unique(orig_strata, return_counts=True)
        stratum_counts = dict(zip(unique_orig.tolist(), counts.tolist()))

        small_strata = [(s, c) for s, c in stratum_counts.items() if c < 2]
        if small_strata:
            raise DimensionError(
                title="Insufficient PSUs per stratum",
                detail=(
                    f"All strata must have at least 2 PSUs. "
                    f"Found {len(small_strata)} strata with fewer."
                ),
                code="INSUFFICIENT_PSU",
                where=where,
                param="stratum",
                expected="≥2 PSUs per stratum",
                got=f"{small_strata[:5]}{'...' if len(small_strata) > 5 else ''}",
                hint="Combine small strata or check design specification.",
            )

        if _method == "brr":
            odd_strata = [(s, c) for s, c in stratum_counts.items() if c % 2 == 1]
            if odd_strata:
                raise DimensionError(
                    title="Odd PSU counts for BRR",
                    detail=(
                        f"BRR requires even number of PSUs per stratum. "
                        f"Found {len(odd_strata)} strata with odd counts."
                    ),
                    code="ODD_PSU_COUNT",
                    where=where,
                    param="stratum",
                    expected="Even PSU count per stratum",
                    got=f"{odd_strata[:5]}{'...' if len(odd_strata) > 5 else ''}",
                    hint="Use method='jk2' which allows 2-3 PSUs per stratum.",
                )

        var_strata = np.empty(n_psus, dtype=np.int64)
        rng = np.random.default_rng(rstate) if shuffle and not order_cols else None
        var_stratum_counter = 0

        for orig_str in unique_orig:
            mask = orig_strata == orig_str
            indices = np.where(mask)[0]
            n_in_stratum = len(indices)

            if rng is not None:
                indices = indices.copy()
                rng.shuffle(indices)

            if method == "brr":
                for i in range(0, n_in_stratum, 2):
                    var_strata[indices[i]] = var_stratum_counter
                    var_strata[indices[i + 1]] = var_stratum_counter
                    var_stratum_counter += 1
            else:  # jk2
                if n_in_stratum % 2 == 1:
                    for i in range(0, n_in_stratum - 3, 2):
                        var_strata[indices[i]] = var_stratum_counter
                        var_strata[indices[i + 1]] = var_stratum_counter
                        var_stratum_counter += 1
                    var_strata[indices[-3]] = var_stratum_counter
                    var_strata[indices[-2]] = var_stratum_counter
                    var_strata[indices[-1]] = var_stratum_counter
                    var_stratum_counter += 1
                else:
                    for i in range(0, n_in_stratum, 2):
                        var_strata[indices[i]] = var_stratum_counter
                        var_strata[indices[i + 1]] = var_stratum_counter
                        var_stratum_counter += 1

        psu_to_var_stratum = dict(zip(psu_list, var_strata))
        psu_vec = df[psu_col].to_list()
        obs_var_strata = np.array([psu_to_var_stratum[p] for p in psu_vec], dtype=np.int64)

        self._sample._data = df.with_columns(pl.Series(name=into, values=obs_var_strata))
        self._sample._design = self._sample._design.update(stratum=into)

        return self._sample

    # ---------- BRR ----------
    def create_brr_wgts(
        self,
        n_reps: int | None = None,
        *,
        rep_prefix: str | None = None,
        fay_coef: float = 0.0,
        rstate: int | None = None,
        drop_nulls: bool = False,
    ) -> Any:
        df = self._sample._data
        design = self._sample._design

        if design.stratum is None:
            raise ValueError("BRR requires `stratum` in Design (got stratum=None).")
        if design.psu is None:
            raise ValueError("BRR requires `psu` in Design (got psu=None).")

        if drop_nulls:
            needed = list(
                {c for c in [design.wgt, design.stratum, design.psu] if isinstance(c, str)}
            )
            data = drop_missing(df=df, cols=needed, treat_infinite_as_missing=True)
        else:
            data = df

        main_weights = _to_float_array(data, design.wgt, len(data))
        stratum_int = _to_int_array(data, design.stratum)
        psu_int = _to_int_array(data, design.psu)

        assert rust_create_brr_wgts is not None  # noqa: S101
        rep_mat, df_val = rust_create_brr_wgts(
            main_weights,
            stratum_int,
            psu_int,
            n_reps,
            fay_coef,
            rstate,
        )
        n_reps_actual = rep_mat.shape[1]

        rep_prefix = rep_prefix or "svy_brr_wgt"
        rep_cols = _name_rep_cols(rep_prefix, n_reps_actual)
        rep_dicts = {col: rep_mat[:, i] for i, col in enumerate(rep_cols)}
        self._sample._data = data.with_columns(
            [pl.Series(name=col, values=vals) for col, vals in rep_dicts.items()]
        )

        self._sample._design = self._sample._design.fill_missing(
            rep_wgts=RepWeights(
                method=EstimationMethod.BRR,
                prefix=rep_prefix,
                n_reps=n_reps_actual,
                fay_coef=fay_coef,
                df=df_val,
            )
        )

        return self._sample

    # ---------- Jackknife ----------
    def create_jk_wgts(
        self,
        *,
        paired: bool = False,
        rep_prefix: str | None = None,
        rstate: int | None = None,
        drop_nulls: bool = False,
    ) -> Any:
        """
        Create Jackknife replicate weights.

        Args:
            paired: If False, use JK1/JKn (delete-one-PSU).
                    If True, use JK2 (paired jackknife).
            rep_prefix: Prefix for replicate weight column names.
            rstate: Random seed for JK2 PSU selection.
            drop_nulls: Drop rows with missing values in design columns.
        """
        df = self._sample._data
        design = self._sample._design

        if design.psu is None:
            raise ValueError("Jackknife requires at least psu; stratum optional.")

        if drop_nulls:
            needed = list(
                {c for c in [design.wgt, design.stratum, design.psu] if isinstance(c, str)}
            )
            data = drop_missing(df=df, cols=needed, treat_infinite_as_missing=True)
        else:
            data = df

        main_weights = _to_float_array(data, design.wgt, len(data))
        psu_int = _to_int_array(data, design.psu)
        stratum_int = _to_int_array(data, design.stratum)

        assert rust_create_jk_wgts is not None  # noqa: S101
        rep_mat, df_val = rust_create_jk_wgts(
            main_weights,
            psu_int,
            stratum_int,
            paired,
            rstate,
        )
        n_reps = rep_mat.shape[1]

        rep_prefix = rep_prefix or ("svy_jk2_wgt" if paired else "svy_jkn_wgt")
        rep_cols = _name_rep_cols(rep_prefix, n_reps)
        rep_dicts = {col: rep_mat[:, i] for i, col in enumerate(rep_cols)}
        self._sample._data = data.with_columns(
            [pl.Series(name=col, values=vals) for col, vals in rep_dicts.items()]
        )

        self._sample._design = self._sample._design.fill_missing(
            rep_wgts=RepWeights(
                method=EstimationMethod.JACKKNIFE,
                prefix=rep_prefix,
                n_reps=n_reps,
                df=df_val,
            )
        )

        return self._sample

    # ---------- Bootstrap ----------
    def create_bs_wgts(
        self,
        n_reps: int = 500,
        *,
        rep_prefix: str | None = None,
        drop_nulls: bool = False,
        rstate: RandomState = None,
    ) -> Any:
        df = self._sample._data
        design = self._sample._design

        if design.psu is None:
            raise ValueError("Bootstrap requires at least psu; stratum optional.")
        if n_reps is None:
            raise ValueError("n_reps must be specified for Bootstrap.")

        if drop_nulls:
            needed = list(
                {c for c in [design.wgt, design.stratum, design.psu] if isinstance(c, str)}
            )
            data = drop_missing(df=df, cols=needed, treat_infinite_as_missing=True)
        else:
            data = df

        main_weights = _to_float_array(data, design.wgt, len(data))
        psu_int = _to_int_array(data, design.psu)
        stratum_int = _to_int_array(data, design.stratum)

        rng = resolve_random_state(rstate)
        seed = (
            int(rng.integers(0, 2**63 - 1))
            if hasattr(rng, "integers")
            else int(rng.randint(0, 2**31 - 1))
        )

        assert rust_create_bootstrap_wgts is not None  # noqa: S101
        rep_mat, df_val = rust_create_bootstrap_wgts(
            main_weights,
            psu_int,
            n_reps,
            stratum_int,
            seed,
        )

        rep_prefix = rep_prefix or "svy_boot_wgt"
        rep_cols = _name_rep_cols(rep_prefix, n_reps)
        rep_dicts = {col: rep_mat[:, i] for i, col in enumerate(rep_cols)}
        self._sample._data = data.with_columns(
            [pl.Series(name=col, values=vals) for col, vals in rep_dicts.items()]
        )

        self._sample._design = self._sample._design.update_rep_weights(
            method=EstimationMethod.BOOTSTRAP,
            prefix=rep_prefix,
            n_reps=n_reps,
            df=df_val,
        )

        return self._sample

    # ---------- SDR ----------
    def create_sdr_wgts(
        self,
        n_reps: int = 4,
        *,
        rep_prefix: str | None = None,
        order_col: str | None = None,
        drop_nulls: bool = False,
    ) -> Any:
        """
        Create SDR (Successive Difference Replication) weights.

        Args:
            n_reps: Number of replicates (typically 4 or more).
            rep_prefix: Prefix for replicate weight column names.
            order_col: Column specifying sort order within strata.
            drop_nulls: Drop rows with missing values in design columns.
        """
        df = self._sample._data
        design = self._sample._design

        if n_reps < 2:
            raise ValueError("SDR requires at least 2 replicates.")

        if drop_nulls:
            needed = list({c for c in [design.wgt, design.stratum] if isinstance(c, str)})
            if order_col:
                needed.append(order_col)
            data = drop_missing(df=df, cols=needed, treat_infinite_as_missing=True)
        else:
            data = df

        main_weights = _to_float_array(data, design.wgt, len(data))
        stratum_int = _to_int_array(data, design.stratum)

        # Order array — single column only
        order_int: np.ndarray | None = None
        if order_col and order_col in data.columns:
            order_int = _to_int_array(data, order_col)

        assert rust_create_sdr_wgts is not None  # noqa: S101
        rep_mat, df_val = rust_create_sdr_wgts(
            main_weights,
            n_reps,
            stratum_int,
            order_int,
        )

        rep_prefix = rep_prefix or "svy_sdr_wgt"
        rep_cols = _name_rep_cols(rep_prefix, n_reps)
        rep_dicts = {col: rep_mat[:, i] for i, col in enumerate(rep_cols)}
        self._sample._data = data.with_columns(
            [pl.Series(name=col, values=vals) for col, vals in rep_dicts.items()]
        )

        self._sample._design = self._sample._design.update_rep_weights(
            method=EstimationMethod.SDR,
            prefix=rep_prefix,
            n_reps=n_reps,
            df=df_val,
        )

        return self._sample

    # ------------------------------------------------------------------ #
    # Reweighting: Adjustment (Non-Response / Within-Class)
    # ------------------------------------------------------------------ #

    def adjust(
        self,
        resp_status: str,
        by: Feature | None,
        *,
        resp_mapping: DomainScalarMap | None = None,
        wgt_name: str | None = None,
        replace: bool = False,
        rep_wgts_prefix: str | None = None,
        ignore_reps: bool = False,
        unknown_to_inelig: bool = True,
        update_design_wgts: bool = True,
        keep_resp: bool = True,
    ) -> Any:
        df = self._sample._data
        design = self._sample._design

        if design.wgt is None:
            raise TypeError("Sample weight is None! Set design.wgt before calling adjust().")
        wgt = design.wgt
        if wgt not in df.columns:
            raise KeyError(f"Weight column {wgt!r} not found in data.")
        if not isinstance(resp_status, str) or resp_status not in df.columns:
            raise TypeError(
                f"`resp_status` must be a string representing an existing column name. "
                f"Got: {resp_status!r}"
            )
        if replace and wgt_name is not None:
            raise TypeError("Pass either replace=True or wgt_name=..., not both.")

        _, adj_class_arr = _normalize_by_term(
            df, by, param_name="by", where="Sample.weighting.adjust"
        )

        wgt_arr = df.get_column(wgt).to_numpy()
        resp_status_arr = df.get_column(resp_status).to_numpy()

        adj_wgt_arr = _adjust_nr(
            wgts=wgt_arr,
            adj_class=adj_class_arr,
            resp_status=resp_status_arr,
            resp_mapping=resp_mapping,
            unknown_to_inelig=unknown_to_inelig,
        )

        existing_cols = set(df.columns)

        if replace:
            target_wgt = wgt
            df = df.with_columns(pl.Series(target_wgt, adj_wgt_arr))
        else:
            target_wgt = _unique_name(
                f"{SVY_ADJ_PREFIX}{wgt}" if wgt_name is None else wgt_name, existing_cols
            )
            df = df.with_columns(pl.Series(target_wgt, adj_wgt_arr))
            existing_cols.add(target_wgt)

        if update_design_wgts:
            self._sample._design = self._sample._design.update(wgt=target_wgt)

        if not ignore_reps and design.rep_wgts is not None:
            rep_cols = design.rep_wgts.columns
            wgts_arr = df.select(rep_cols).to_numpy()

            if adj_class_arr is not None:
                unique_classes = np.unique(adj_class_arr)
                class_to_idx = {cls: idx for idx, cls in enumerate(unique_classes)}
                adj_class_indices = np.array(
                    [class_to_idx[cls] for cls in adj_class_arr], dtype=np.int64
                )
            else:
                adj_class_indices = np.zeros(len(wgt_arr), dtype=np.int64)

            resp_codes = np.zeros(len(resp_status_arr), dtype=np.int64)
            resp_lower = np.char.lower(resp_status_arr.astype(str))
            CANONICAL_TO_INT = {"rr": 0, "nr": 1, "in": 2, "uk": 3}

            if resp_mapping is not None:
                for canonical_key, data_label in resp_mapping.items():
                    key_lower = str(canonical_key).lower()
                    if key_lower not in CANONICAL_TO_INT:
                        raise ValueError(
                            f"Invalid canonical key '{canonical_key}'. "
                            f"Must be one of: rr, nr, in, uk"
                        )
                    int_code = CANONICAL_TO_INT[key_lower]
                    resp_codes[resp_lower == str(data_label).lower()] = int_code
            else:
                resp_codes[resp_lower == "rr"] = 0
                resp_codes[resp_lower == "nr"] = 1
                resp_codes[resp_lower == "in"] = 2
                resp_codes[resp_lower == "uk"] = 3

            assert rust_adjust_nr is not None  # noqa: S101
            adj_wgts_arr = rust_adjust_nr(
                wgts_arr,
                adj_class_indices,
                resp_codes,
                unknown_to_inelig,
            )

            wgts_df = pl.DataFrame(adj_wgts_arr)
            n_reps = len(rep_cols)

            if rep_wgts_prefix is not None:
                final_prefix = rep_wgts_prefix
            else:
                final_prefix = f"{SVY_ADJ_PREFIX}{design.rep_wgts.prefix}"

            new_rep_names = [
                _unique_name(f"{final_prefix}{i}", existing_cols) for i in range(1, n_reps + 1)
            ]
            wgts_df.columns = new_rep_names

            self._sample._data = df.hstack(wgts_df)
            df = self._sample._data

            if update_design_wgts:
                self._sample._design = self._sample._design.update_rep_weights(
                    method=design.rep_wgts.method,
                    prefix=final_prefix,
                    n_reps=n_reps,
                    fay_coef=design.rep_wgts.fay_coef,
                    df=design.rep_wgts.df,
                )

        self._sample._data = df

        if keep_resp:

            def _resp_mask_expr(col: str, mapping: dict | None) -> pl.Expr:
                if mapping is None:
                    return pl.col(col) == "rr"
                rr_codes = mapping.get("rr", "rr")
                if isinstance(rr_codes, (list, tuple, set, np.ndarray)):
                    return pl.col(col).is_in(list(rr_codes))
                else:
                    return pl.col(col) == rr_codes

            mask = _resp_mask_expr(resp_status, resp_mapping)
            self._sample._data = self._sample._data.filter(mask)

        return self._sample

    # ------------------------------------------------------------------ #
    # Reweighting: Normalization
    # ------------------------------------------------------------------ #

    def normalize(
        self,
        controls: DomainScalarMap | Number | None = None,
        *,
        by: Feature | None = None,
        wgt_name: str | None = None,
        replace: bool = False,
        rep_wgts_prefix: str | None = None,
        ignore_reps: bool = False,
        update_design_wgts: bool = True,
    ) -> Any:
        df = self._sample._data
        design = self._sample._design

        if design.wgt is None:
            raise TypeError("Sample weight is None! Set design.wgt before calling normalize().")
        wgt = design.wgt
        if wgt not in df.columns:
            raise KeyError(f"Weight column {wgt!r} not found in data.")
        if replace and wgt_name is not None:
            raise TypeError("Pass either replace=True or wgt_name=..., not both.")

        _, by_arr = _normalize_by_term(df, by, param_name="by", where="Sample.weighting.normalize")

        wgt_arr = df.get_column(wgt).to_numpy()
        existing_cols = set(df.columns)

        samp_norm, _, _ = _normalize(wgt=wgt_arr, control=controls, by_arr=by_arr)

        if replace:
            target_wgt = wgt
            df = df.with_columns(pl.Series(name=target_wgt, values=samp_norm))
        else:
            target_wgt = _unique_name(wgt_name or f"{SVY_NORM_PREFIX}{wgt}", existing_cols)
            df = df.with_columns(pl.Series(name=target_wgt, values=samp_norm))
            existing_cols.add(target_wgt)

        if update_design_wgts:
            self._sample._design = self._sample._design.update(wgt=target_wgt)

        if not ignore_reps and design.rep_wgts is not None:
            rep_cols = design.rep_wgts.columns

            if rep_cols:
                wgts_arr = df.select(rep_cols).to_numpy()

                if by_arr is None:
                    assert rust_normalize is not None  # noqa: S101
                    adj_wgts_arr = rust_normalize(wgts_arr, None, None)
                else:
                    raise NotImplementedError(
                        "Grouped normalization on replicate weights not yet supported. "
                        "Use by=None for global normalization."
                    )

                if rep_wgts_prefix is None:
                    final_prefix = f"{SVY_NORM_PREFIX}{design.rep_wgts.prefix}"
                else:
                    final_prefix = rep_wgts_prefix

                n_reps = len(rep_cols)
                new_rep_names = [f"{final_prefix}{i}" for i in range(1, n_reps + 1)]

                wgts_df = pl.DataFrame(adj_wgts_arr, schema=new_rep_names)
                self._sample._data = df.hstack(wgts_df)
                df = self._sample._data

                if update_design_wgts:
                    self._sample._design = self._sample._design.update_rep_weights(
                        method=design.rep_wgts.method,
                        prefix=final_prefix,
                        n_reps=n_reps,
                        fay_coef=design.rep_wgts.fay_coef,
                        df=design.rep_wgts.df,
                    )

        self._sample._data = df
        return self._sample

    # ------------------------------------------------------------------ #
    # Reweighting: Poststratification
    # ------------------------------------------------------------------ #

    def poststratify(
        self,
        controls: DomainScalarMap | Number | None = None,
        *,
        factors: DomainScalarMap | Number | None = None,
        by: Feature | None = None,
        wgt_name: str | None = None,
        replace: bool = False,
        rep_wgts_prefix: str | None = None,
        ignore_reps: bool = False,
        update_design_wgts: bool = True,
    ) -> Any:
        if controls is None and factors is None:
            raise AssertionError("control or factor must be specified.")

        df = self._sample._data
        design = self._sample._design

        if design.wgt is None:
            raise TypeError("Sample weight is None! Set design.wgt before calling poststratify().")
        wgt = design.wgt
        if wgt not in df.columns:
            raise KeyError(f"Weight column {wgt!r} not found in data.")
        if replace and wgt_name is not None:
            raise ValueError("Pass either replace=True or wgt_name=..., not both.")

        _, by_arr = _normalize_by_term(
            df, by, param_name="by", where="Sample.weighting.poststratify"
        )

        wgt_arr = df.get_column(wgt).to_numpy()
        existing_cols = set(df.columns)

        ps_wgt = _poststratify(
            wgt=wgt_arr,
            control=controls,
            factor=factors,
            by_arr=by_arr,
        )

        if replace:
            target_wgt = wgt
            df = df.with_columns(pl.Series(name=target_wgt, values=ps_wgt))
        else:
            target_wgt = _unique_name(wgt_name or f"{SVY_PS_PREFIX}{wgt}", existing_cols)
            df = df.with_columns(pl.Series(name=target_wgt, values=ps_wgt))
            existing_cols.add(target_wgt)

        if update_design_wgts:
            self._sample._design = self._sample._design.update(wgt=target_wgt)

        if not ignore_reps and design.rep_wgts is not None:
            rep_cols = design.rep_wgts.columns

            if rep_cols:
                wgts_arr = df.select(rep_cols).to_numpy()

                if by_arr is None:
                    by_indices = np.zeros(len(wgt_arr), dtype=np.int64)

                    if isinstance(controls, dict):
                        if len(controls) != 1:
                            raise ValueError(
                                "When by=None (single stratum), controls dict must have "
                                "exactly one key"
                            )
                        control_arr = np.array(
                            [float(next(iter(controls.values())))], dtype=np.float64
                        )
                    elif isinstance(controls, (int, float)):
                        control_arr = np.array([float(controls)], dtype=np.float64)
                    elif factors is not None:
                        if isinstance(factors, dict):
                            if len(factors) != 1:
                                raise ValueError(
                                    "When by=None (single stratum), factors dict must have "
                                    "exactly one key"
                                )
                            factor_val = float(next(iter(factors.values())))
                            control_arr = np.array([factor_val * wgt_arr.sum()], dtype=np.float64)
                        else:
                            control_arr = np.array(
                                [float(factors) * wgt_arr.sum()], dtype=np.float64
                            )
                    else:
                        raise ValueError(
                            "Must provide controls or factors for post-stratification"
                        )

                    assert rust_poststratify is not None  # noqa: S101
                    adj_wgts_arr = rust_poststratify(wgts_arr, by_indices, control_arr)

                elif isinstance(controls, dict):
                    unique_vals = np.unique(by_arr)
                    val_to_idx = {val: idx for idx, val in enumerate(unique_vals)}
                    by_indices = np.array([val_to_idx[val] for val in by_arr], dtype=np.int64)
                    control_arr = np.array(
                        [float(controls[val]) for val in unique_vals], dtype=np.float64
                    )

                    assert rust_poststratify is not None  # noqa: S101
                    adj_wgts_arr = rust_poststratify(wgts_arr, by_indices, control_arr)

                elif isinstance(controls, (int, float)):
                    unique_vals = np.unique(by_arr)
                    val_to_idx = {val: idx for idx, val in enumerate(unique_vals)}
                    by_indices = np.array([val_to_idx[val] for val in by_arr], dtype=np.int64)
                    n_strata = len(unique_vals)
                    control_arr = np.zeros(n_strata, dtype=np.float64)
                    total_weight = wgt_arr.sum()
                    for idx, val in enumerate(unique_vals):
                        mask = by_arr == val
                        stratum_weight = wgt_arr[mask].sum()
                        control_arr[idx] = float(controls) * (stratum_weight / total_weight)

                    assert rust_poststratify is not None  # noqa: S101
                    adj_wgts_arr = rust_poststratify(wgts_arr, by_indices, control_arr)

                elif factors is not None:
                    unique_vals = np.unique(by_arr)
                    val_to_idx = {val: idx for idx, val in enumerate(unique_vals)}
                    by_indices = np.array([val_to_idx[val] for val in by_arr], dtype=np.int64)

                    if isinstance(factors, dict):
                        control_arr = np.zeros(len(unique_vals), dtype=np.float64)
                        for idx, val in enumerate(unique_vals):
                            mask = by_arr == val
                            stratum_weight = wgt_arr[mask].sum()
                            control_arr[idx] = float(factors[val]) * stratum_weight
                    else:
                        control_arr = np.zeros(len(unique_vals), dtype=np.float64)
                        for idx, val in enumerate(unique_vals):
                            mask = by_arr == val
                            stratum_weight = wgt_arr[mask].sum()
                            control_arr[idx] = float(factors) * stratum_weight

                    assert rust_poststratify is not None  # noqa: S101
                    adj_wgts_arr = rust_poststratify(wgts_arr, by_indices, control_arr)

                else:
                    raise ValueError("Must provide controls or factors for post-stratification")

                if rep_wgts_prefix is None:
                    final_prefix = f"{SVY_PS_PREFIX}{design.rep_wgts.prefix}"
                else:
                    final_prefix = rep_wgts_prefix

                n_reps = len(rep_cols)
                new_rep_names = [f"{final_prefix}{i}" for i in range(1, n_reps + 1)]

                wgts_df = pl.DataFrame(adj_wgts_arr, schema=new_rep_names)
                self._sample._data = df.hstack(wgts_df)
                df = self._sample._data

                if update_design_wgts:
                    self._sample._design = self._sample._design.update_rep_weights(
                        method=design.rep_wgts.method,
                        prefix=final_prefix,
                        n_reps=n_reps,
                        fay_coef=design.rep_wgts.fay_coef,
                        df=design.rep_wgts.df,
                    )

        self._sample._data = df
        return self._sample

    # ------------------------------------------------------------------ #
    # Raking
    # ------------------------------------------------------------------ #

    def controls_margins_template(
        self,
        *,
        margins: Mapping[str, str],
        cat_na: str = "level",
        na_label: str = "__NA__",
    ) -> dict[str, dict[Category, float]]:
        df: pl.DataFrame = self._sample.data
        where = "Sample.weighting.controls_margins_template"

        for mname, col in margins.items():
            if not isinstance(mname, str) or not isinstance(col, str):
                raise TypeError(
                    f"{where}: margins must be {{str -> str}}; got {type(mname)!r}->{type(col)!r}"
                )
            if col not in df.columns:
                raise KeyError(f"{where}: column {col!r} (for margin {mname!r}) not found.")

        result: dict[str, dict[Category, float]] = {}

        for mname, col in margins.items():
            s = df.get_column(col)

            if cat_na not in ("error", "level"):
                raise ValueError(f"{where}: cat_na must be 'error' or 'level'.")

            if cat_na == "error":
                if s.is_null().any():
                    raise DimensionError(
                        title="Missing values in margin column",
                        detail=f"Nulls found in {col!r}. Choose cat_na='level' or fix data.",
                        code="MARGIN_NA",
                        where=where,
                        param=col,
                        hint="Use cat_na='level' to include a missing category.",
                    )
                s_norm = s.cast(pl.Utf8)
            else:
                s_norm = s.cast(pl.Utf8).fill_null(na_label)

            cats = pl.Series("__cats__", s_norm.unique().to_list(), dtype=pl.Utf8).to_list()
            cats_sorted = sorted(cats, key=_num_sort_key_label)
            result[mname] = {lab: np.nan for lab in cats_sorted}

        return result

    def rake(
        self,
        *,
        controls: ControlsType | None = None,
        factors: ControlsType | None = None,
        wgt_name: str | None = None,
        replace: bool = False,
        rep_wgts_prefix: str | None = None,
        ignore_reps: bool = False,
        ll_bound: float | None = None,
        up_bound: float | None = None,
        tol: float = 1e-4,
        max_iter: int = 100,
        display_iter: bool = False,
        update_design_wgts: bool = True,
    ) -> "Sample":
        df = self._sample._data
        design = self._sample._design

        if design.wgt is None:
            raise TypeError("Sample weight is None! Set design.wgt before calling rake().")
        wgt = design.wgt
        if wgt not in df.columns:
            raise KeyError(f"Weight column {wgt!r} not found in data.")
        if replace and wgt_name is not None:
            raise TypeError("Pass either replace=True or wgt_name=..., not both.")

        controls_norm: ControlsType | None = _normalize_controls_like(x=controls)
        factors_norm: ControlsType | None = _normalize_controls_like(x=factors)

        if controls_norm is None and factors_norm is None:
            raise ValueError("Either 'control' or 'factor' must be specified.")
        if controls_norm is not None and factors_norm is not None:
            raise ValueError("Provide exactly one of 'control' or 'factor'.")

        if ll_bound is not None and up_bound is not None and ll_bound > up_bound:
            raise ValueError("ll_bound cannot be greater than up_bound.")

        rake_cols = (
            list(controls_norm.keys()) if controls_norm is not None else list(factors_norm.keys())  # type: ignore[union-attr]
        )
        if not rake_cols:
            raise ValueError("No raking columns provided in control/factor keys.")

        processed: dict[str, np.ndarray] = {}
        w0 = df.get_column(wgt).to_numpy()

        for col in rake_cols:
            if not isinstance(col, str) or col not in df.columns:
                raise KeyError(f"Raking column {col!r} not found in data.")
            s = df.get_column(col)
            if s.len() != w0.size:
                raise ValueError(f"Raking column {col!r} length mismatch.")
            if s.null_count() > 0:
                raise ValueError(f"Raking column {col!r} contains nulls.")
            processed[col] = s.to_numpy()

        control_final: ControlsType = controls_norm or _calculate_controls_from_factors(
            wgts=w0,
            margins=processed,
            factors=cast(ControlsType, factors_norm),
        )

        missing = [m for m in processed if m not in control_final]
        extra = [m for m in control_final if m not in processed]
        if missing or extra:
            raise ValueError(
                "Control totals must be provided for all specified raking columns "
                f"(missing={missing or []}, extra={extra or []})."
            )

        for col_name, totals in control_final.items():
            if not isinstance(totals, Mapping) or not totals:
                raise ValueError(f"Control for column {col_name!r} must be a non-empty mapping.")
            vals = np.array(list(totals.values()), dtype=float)
            if not np.all(np.isfinite(vals)) or np.any(vals < 0):
                raise ValueError(
                    f"Control totals for {col_name!r} must be finite and non-negative."
                )
            if np.all(vals == 0):
                raise ValueError(f"All-zero control totals for {col_name!r} not allowed.")

        raked_w = _rake(
            wgt=w0,
            margins=processed,
            control=control_final,
            ll_bound=ll_bound,
            up_bound=up_bound,
            tol=tol,
            max_iter=max_iter,
            display_iter=display_iter,
        )

        existing_cols = set(df.columns)
        if replace:
            target_wgt = wgt
            df = df.with_columns(pl.Series(name=target_wgt, values=raked_w))
        else:
            target_wgt = _unique_name(wgt_name or f"{SVY_RAKED_PREFIX}{wgt}", existing_cols)
            df = df.with_columns(pl.Series(name=target_wgt, values=raked_w))
            existing_cols.add(target_wgt)

        if update_design_wgts:
            self._sample._design = self._sample._design.update(wgt=target_wgt)

        if not ignore_reps and design.rep_wgts is not None:
            rep_cols = design.rep_wgts.columns

            if rep_cols:
                if rep_wgts_prefix is None:
                    final_prefix = f"{SVY_RAKED_PREFIX}{design.rep_wgts.prefix}"
                else:
                    final_prefix = rep_wgts_prefix

                n_reps = len(rep_cols)
                wgts_arr = df.select(rep_cols).to_numpy()

                margin_indices = []
                margin_targets = []

                for col in rake_cols:
                    cats_sorted = sorted(control_final[col].keys(), key=_num_sort_key_label)
                    cat_to_idx = {cat: idx for idx, cat in enumerate(cats_sorted)}
                    indices = np.array([cat_to_idx[val] for val in processed[col]], dtype=np.int64)
                    targets = np.array(
                        [float(control_final[col][cat]) for cat in cats_sorted],
                        dtype=np.float64,
                    )
                    margin_indices.append(indices)
                    margin_targets.append(targets)

                assert rust_rake is not None  # noqa: S101
                raked_reps = rust_rake(
                    wgts_arr,
                    margin_indices,
                    margin_targets,
                    ll_bound,
                    up_bound,
                    tol,
                    max_iter,
                )

                new_rep_names = [f"{final_prefix}{i}" for i in range(1, n_reps + 1)]
                wgts_df = pl.DataFrame(raked_reps, schema=new_rep_names)
                self._sample._data = df.hstack(wgts_df)
                df = self._sample._data

                if update_design_wgts:
                    self._sample._design = self._sample._design.update_rep_weights(
                        method=design.rep_wgts.method,
                        prefix=final_prefix,
                        n_reps=n_reps,
                        fay_coef=design.rep_wgts.fay_coef,
                        df=design.rep_wgts.df,
                    )

        self._sample._data = df
        return self._sample

    # ------------------------------------------------------------------ #
    # Calibration
    # ------------------------------------------------------------------ #

    def control_aux_template(
        self,
        *,
        x: Sequence[Feature],
        by: Feature | None = None,
        by_na: Literal["error", "level", "drop"] = "error",
        na_label: str = "__NA__",
    ) -> dict[Category, Number] | dict[Category, dict[Category, Number]]:
        _, shape = self.build_aux_matrix(x=x, by=by, by_na=by_na, na_label=na_label)
        return shape

    def build_aux_matrix(
        self,
        *,
        x: Sequence[Feature],
        by: Feature | None = None,
        by_na: Literal["error", "level", "drop"] = "error",
        na_label: str = "__NA__",
    ) -> tuple[np.ndarray, dict[Category, Number] | dict[Category, dict[Category, Number]]]:
        df: pl.DataFrame = self._sample.data
        where = "Sample.weighting.build_aux_matrix"

        by_n, _ = _normalize_by_term(df, by, param_name="by", where=where)

        if not x:
            raise MethodError.not_applicable(
                where=where, method="aux builder", reason="No terms specified."
            )

        all_exprs = []
        all_labels = []

        for term in x:
            t_exprs, t_labs = _expand_term(term, df, where)
            all_exprs.extend(t_exprs)
            all_labels.extend(t_labs)

        if not all_exprs:
            raise MethodError.not_applicable(
                where=where, method="aux builder", reason="Terms produced no columns."
            )

        final_exprs = [e.alias(f"__aux_col_{i}") for i, e in enumerate(all_exprs)]

        if by_n is None:
            X_df = df.select(final_exprs)
            X = X_df.to_numpy()
            return X, {lab: np.nan for lab in all_labels}

        if by_na == "error":
            if (
                df.select([pl.col(c).is_null().any() for c in by_n])
                .select(pl.any_horizontal(pl.all()))
                .item()
            ):
                raise DimensionError(
                    title="Missing values in `by` columns",
                    detail=f"Nulls found in {by_n}.",
                    code="BY_NA",
                    where=where,
                    param="by",
                    hint="Use by_na='level' or 'drop', or fix data.",
                )
            by_exprs = [pl.col(c) for c in by_n]
        elif by_na == "level":
            by_exprs = [pl.col(c).fill_null(na_label) for c in by_n]
        elif by_na == "drop":
            filter_expr = pl.all_horizontal([pl.col(c).is_not_null() for c in by_n])
            df = df.filter(filter_expr)
            by_exprs = [pl.col(c) for c in by_n]
        else:
            raise ValueError("by_na must be 'error', 'level', or 'drop'.")

        combined_df = df.select(final_exprs + by_exprs)
        n_x_cols = len(final_exprs)
        X = combined_df[:, :n_x_cols].to_numpy()
        by_arr = combined_df[:, n_x_cols:].to_numpy()

        _keys_raw = [row[0] if by_arr.shape[1] == 1 else tuple(row.tolist()) for row in by_arr]
        keys: list[Category] = _keys_raw  # type: ignore[assignment]

        try:
            uniq_keys = sorted(list(set(keys)))
        except TypeError:
            uniq_keys = list(set(keys))

        inner_template = {lab: np.nan for lab in all_labels}
        return X, {k: inner_template.copy() for k in uniq_keys}

    def calibrate(
        self,
        *,
        controls: dict[Feature, Any],
        by: Feature | None = None,
        scale: Number | list[Number] | np.ndarray = 1.0,
        bounded: bool = False,
        additive: bool = False,
        wgt_name: str | None = None,
        replace: bool = False,
        update_design_wgts: bool = True,
        rep_wgts_prefix: str | None = None,
        ignore_reps: bool = False,
    ) -> Any:
        where = "Sample.weighting.calibrate"

        if not isinstance(controls, dict) or not controls:
            raise MethodError.not_applicable(
                where=where,
                method="calibrate",
                reason="`controls` must be a non-empty dictionary.",
                param="controls",
            )

        terms: list[Feature]
        is_global = by is None

        if is_global:
            terms = list(controls.keys())
        else:
            first_domain_val = next(iter(controls.values()))
            if not isinstance(first_domain_val, dict):
                raise TypeError(
                    f"When `by` is used, `controls` values must be dictionaries of "
                    f"targets. Got {type(first_domain_val)}."
                )
            terms = list(first_domain_val.keys())

        X, shape_template = self.build_aux_matrix(x=terms, by=by)

        if is_global:
            x_labels = list(shape_template.keys())
        else:
            x_labels = list(next(iter(shape_template.values())).keys())

        final_control_arg: Any

        if is_global:
            flat_targets = []
            cursor = 0
            for term in terms:
                _, term_labs = _expand_term(term, self._sample.data, where)
                k = len(term_labs)
                segment_labels = x_labels[cursor : cursor + k]
                if segment_labels != term_labs:
                    raise RuntimeError("Internal error: Design matrix label alignment mismatch.")
                spec = controls[term]
                vals = _match_term_targets(segment_labels, spec, str(term))
                flat_targets.extend(vals)
                cursor += k
            final_control_arg = np.array(flat_targets, dtype=float)
        else:
            final_control_arg = {}
            expected_domains = set(shape_template.keys())
            provided_domains = set(controls.keys())
            missing = expected_domains - provided_domains
            extra = provided_domains - expected_domains
            if missing or extra:
                raise ValueError(f"Domain mismatch in controls. Missing={missing}, Extra={extra}")

            for domain, domain_specs in controls.items():
                flat_targets = []
                cursor = 0
                for term in terms:
                    if term not in domain_specs:
                        raise ValueError(
                            f"Missing specification for term '{term}' in domain '{domain}'."
                        )
                    _, term_labs = _expand_term(term, self._sample.data, where)
                    k = len(term_labs)
                    segment_labels = x_labels[cursor : cursor + k]
                    spec = domain_specs[term]
                    vals = _match_term_targets(segment_labels, spec, str(term))
                    flat_targets.extend(vals)
                    cursor += k
                final_control_arg[domain] = dict(zip(x_labels, flat_targets))

        return self.calibrate_matrix(
            aux_vars=X,
            control=final_control_arg,
            by=by,
            scale=scale,
            bounded=bounded,
            additive=additive,
            wgt_name=wgt_name,
            replace=replace,
            update_design_wgts=update_design_wgts,
            labels=x_labels,
            weights_only=False,
            rep_wgts_prefix=rep_wgts_prefix,
            ignore_reps=ignore_reps,
        )

    def calibrate_matrix(
        self,
        *,
        aux_vars: np.ndarray,
        control: Any,
        by: Feature | None = None,
        scale: Number | Sequence[Number] | np.ndarray = 1.0,
        wgt_name: str | None = None,
        replace: bool = False,
        update_design_wgts: bool = True,
        labels: Sequence[Category] | None = None,
        weights_only: bool = False,
        bounded: bool = False,
        additive: bool = False,
        rep_wgts_prefix: str | None = None,
        ignore_reps: bool = False,
    ) -> Any:
        where = "Sample.weighting.calibrate_matrix"
        df: pl.DataFrame = self._sample.data
        design = self._sample._design

        wgt_col = getattr(self._sample.design, "wgt")
        if not isinstance(wgt_col, str) or wgt_col not in df.columns:
            raise MethodError.invalid_choice(
                where=where,
                param="design.wgt",
                got=wgt_col,
                allowed=df.columns,
                hint="Design must reference an existing weight column.",
            )
        w = df.get_column(wgt_col).to_numpy()

        X = np.asarray(aux_vars, dtype=float)
        if X.ndim == 1:
            X = X.reshape(-1, 1)
        if X.shape[0] != df.height:
            raise DimensionError(
                title="Shape mismatch",
                detail=f"aux_vars has {X.shape[0]} rows; data has {df.height}.",
                code="SHAPE_MISMATCH",
                where=where,
                param="aux_vars",
                expected=f"(n,{X.shape[1]}) with n={df.height}",
                got=f"(n={X.shape[0]},{X.shape[1]})",
            )

        _, domain_vec = _normalize_by_term(df, by, param_name="by", where=where)

        if domain_vec is None and isinstance(control, dict) and labels is not None:
            missing_labels = [lbl for lbl in labels if lbl not in control]
            extra_keys = [k for k in control.keys() if k not in labels]
            if missing_labels or extra_keys:
                raise ValueError(
                    f"Control keys mismatch. Missing: {missing_labels}, Extra: {extra_keys}"
                )
            control = np.array([float(control[lbl]) for lbl in labels])

        new_w = _calibrate(
            samp_weight=w,
            aux_vars=X,
            control=control,
            domain=domain_vec,
            scale=scale,
            bounded=bounded,
            additive=additive,
        )

        if weights_only:
            return new_w

        existing_cols = set(df.columns)

        wgt_col_name = wgt_name or f"{SVY_CALIB_PREFIX}{wgt_col}"

        if wgt_col_name in df.columns:
            raise MethodError.invalid_choice(
                where=where,
                param="wgt_name",
                got=wgt_col_name,
                allowed=[c for c in df.columns if c != wgt_col_name],
                hint="Choose a new column name or drop the existing one first.",
            )

        df = df.with_columns(pl.Series(name=wgt_col_name, values=new_w))
        existing_cols.add(wgt_col_name)

        if replace and update_design_wgts:
            if hasattr(design, "update") and callable(getattr(design, "update")):
                self._sample._design = design.update(wgt=wgt_col_name)
            else:
                setattr(design, "wgt", wgt_col_name)

        if not ignore_reps and design.rep_wgts is not None:
            rep_cols = design.rep_wgts.columns

            if rep_cols:
                wgts_arr = df.select(rep_cols).to_numpy()

                if isinstance(scale, (int, float)):
                    scale_arr = np.full(len(w), float(scale), dtype=np.float64)
                else:
                    scale_arr = np.asarray(scale, dtype=np.float64)

                if domain_vec is None:
                    if isinstance(control, dict):
                        if labels is not None:
                            totals_arr = np.array(
                                [float(control[lbl]) for lbl in labels],  # type: ignore[index]
                                dtype=np.float64,
                            )
                        else:
                            totals_arr = np.array(list(control.values()), dtype=np.float64)
                    else:
                        totals_arr = np.asarray(control, dtype=np.float64)

                    n_reps = len(rep_cols)
                    if n_reps >= 10:
                        assert rust_calibrate_parallel is not None  # noqa: S101
                        calib_reps = rust_calibrate_parallel(
                            wgts_arr,
                            X,
                            totals_arr,
                            scale_arr,
                        )
                    else:
                        assert rust_calibrate is not None  # noqa: S101
                        calib_reps = rust_calibrate(
                            wgts_arr,
                            X,
                            totals_arr,
                            scale_arr,
                            additive,
                        )
                else:
                    unique_domains = np.unique(domain_vec)
                    domain_to_idx = {d: idx for idx, d in enumerate(unique_domains)}
                    domain_indices = np.array(
                        [domain_to_idx[d] for d in domain_vec], dtype=np.int64
                    )

                    controls_dict: dict[int, list[float]] = {}
                    for domain in unique_domains:
                        domain_idx = domain_to_idx[domain]
                        if isinstance(control, dict):
                            domain_control = control[domain]
                            if isinstance(domain_control, dict):
                                if labels is not None:
                                    totals = [float(domain_control[lbl]) for lbl in labels]
                                else:
                                    totals = list(domain_control.values())
                            else:
                                totals = list(np.asarray(domain_control, dtype=float))
                        else:
                            totals = list(np.asarray(control, dtype=float))
                        controls_dict[domain_idx] = totals

                    assert rust_calibrate_by_domain is not None  # noqa: S101
                    calib_reps = rust_calibrate_by_domain(
                        wgts_arr,
                        X,
                        domain_indices,
                        controls_dict,
                        scale_arr,
                        additive,
                    )

                if rep_wgts_prefix is None:
                    final_prefix = f"{SVY_CALIB_PREFIX}{design.rep_wgts.prefix}"
                else:
                    final_prefix = rep_wgts_prefix

                n_reps = len(rep_cols)
                new_rep_names = [f"{final_prefix}{i}" for i in range(1, n_reps + 1)]

                wgts_df = pl.DataFrame(calib_reps, schema=new_rep_names)
                df = df.hstack(wgts_df)

                if update_design_wgts:
                    self._sample._design = self._sample._design.update_rep_weights(
                        method=design.rep_wgts.method,
                        prefix=final_prefix,
                        n_reps=n_reps,
                        fay_coef=design.rep_wgts.fay_coef,
                        df=design.rep_wgts.df,
                    )

        self._sample._data = df
        return self._sample



# src/svy/core/enumerations.py
"""Provides the custom enums used throughout the modules."""

from enum import Enum, StrEnum, unique


@unique
class CaseStyle(StrEnum):
    SNAKE = "Snake"
    CAMEL = "Camel"
    PASCAL = "Pascal"
    KEBAB = "Kebab"


@unique
class DistFamily(StrEnum):
    GAUSSIAN = "Gaussian"
    BINOMIAL = "Binomial"
    POISSON = "Poisson"
    GAMMA = "Gamma"
    INVERSE_GAUSSIAN = "InverseGaussian"
    # Phase 2:
    # NEG_BINOMIAL = "Negative Binomial"
    # BETA = "Beta"


# @unique
# class EditMethod(StrEnum):
#     BOTTOM_CODING = "Bottom Code"
#     TOP_CODING = "Top Code"
#     BOTTOM_AND_TOP_CODING = "Top and Bottom Code"
#     RECODING = "Recode"
#     CATEGORIZING = "Categorize"
#     RENAMING = "Rename"
#     CLEANING_NAMES = "Cleaning Names"


@unique
class EstimationMethod(StrEnum):
    # Standard variance estimation (Linearization)
    TAYLOR = "Taylor"

    # Replicate variance estimation methods
    BRR = "BRR"
    BOOTSTRAP = "Bootstrap"
    JACKKNIFE = "Jackknife"
    SDR = "SDR"  # Successive Difference Replication (e.g. used by ACS)


@unique
class FitMethod(StrEnum):
    # OLS = "OLS"
    # WLS = "WLS"
    # FH = "FH"
    ML = "ML"
    REML = "REML"


@unique
class MeanVarMode(StrEnum):
    EQUAL_VAR = "Equal Variances"
    """Assume σ1 = σ2 (common variance)."""

    UNEQUAL_VAR = "Unequal Variances"
    """Allow σ1 ≠ σ2 (Welch-type); supports optimal allocation."""


@unique
class ModelType(StrEnum):
    LINEAR = "Linear"  # Gaussian + Identity
    LOGISTIC = "Logistic"  # Binomial + Logit
    POISSON = "Poisson"  # Poisson + Log
    GAMMA = "Gamma"  # Gamma + Log (or Inverse)


@unique
class MetadataSource(StrEnum):
    """Track where variable metadata originated from."""

    INFERRED = "inferred"  # Auto-detected from data types/values
    SCHEMA = "schema"  # From explicit Schema definition
    QUESTIONNAIRE = "questionnaire"  # From Questionnaire definition
    IMPORTED = "imported"  # From external file (SPSS, Stata, etc.)
    USER = "user"  # Manually set by user at runtime


@unique
class MissingKind(StrEnum):
    """
    Semantic classification of missing values.

    These kinds help distinguish between different reasons for missingness,
    which can inform analysis decisions (e.g., imputation strategy).

    User-generated (typically MNAR):
        DONT_KNOW - Respondent doesn't know the answer
        REFUSED - Respondent refused to answer
        NO_ANSWER - No answer provided (ambiguous cause)

    System/Design-generated (typically MAR/MCAR):
        SYSTEM - System-generated missing (data processing issue)
        SKIPPED - Question skipped due to routing/skip logic
        NOT_APPLICABLE - Question not applicable to respondent
        STRUCTURAL - Missing by study design (e.g., split questionnaire)
    """

    # User-generated missing (typically MNAR)
    DONT_KNOW = "dnk"  # -> is typically MNAR
    REFUSED = "refused"  # -> is typically MNAR
    NO_ANSWER = "no_answer"  # -> could be any of MNAR, MAR, or MCAR

    # System/design-generated missing (typically MAR/MCAR)
    STRUCTURAL = "struct"  # -> is typical MAR e.g. SKIPPED and N/A
    SYSTEM = "system"  # -> is typically MCAR


@unique
class MissingMechanism(StrEnum):
    MCAR = "Completely Missing At Random"
    MAR = "Missing At Random"
    MNAR = "Missing Not At Random"


@unique
class LetterCase(str, Enum):
    LOWER = "lower"
    UPPER = "upper"
    TITLE = "title"
    ORIGINAL = "original"


@unique
class LinkFunction(StrEnum):
    IDENTITY = "identity"
    LOGIT = "logit"
    LOG = "log"
    INVERSE = "inverse"
    INVERSE_SQUARED = "inverse_squared"
    # PROBIT = "probit"  # Requires CDF implementation
    # CLOGLOG = "cloglog"


@unique
class MeasurementType(StrEnum):
    CONTINUOUS = "Numerical Continuous"
    DISCRETE = "Numerical Discrete"
    NOMINAL = "Categorical Nominal"
    ORDINAL = "Categorical Ordinal"
    STRING = "String"
    BOOLEAN = "Boolean"
    DATETIME = "Datetime"


@unique
class OnePropSizeMethod(StrEnum):
    WALD = "Wald"
    FLEISS = "Fleiss"
    WILSON = "Wilson"
    AGRESTI_COULL = "Agresti-Coull"
    CLOPPER_PEARSON = "Clopper-Pearson"


@unique
class PopParam(StrEnum):
    MEAN = "Mean"
    TOTAL = "Total"
    PROP = "Proportion"
    RATIO = "Ratio"
    MEDIAN = "Median"


@unique
class PPSMethod(StrEnum):
    BREWER = "PPS Brewer"
    # HV = "PPS Hanurav-Vijayan"
    MURPHY = "PPS Murphy"
    RS = "PPS Rao-Sampford"
    SYS = "PPS Systematic"
    WR = "PPS with replacement"


@unique
class PropVarMode(StrEnum):
    ALT_PROPS = "Alternative Proportions"
    """Variance under alternative: p1(1-p1), p2(1-p2)."""

    POOLED_PROP = "Pooled Proportion"
    """Variance under null with pooled p; conservative/classic."""

    MAX_VAR = "Maximum Variance"
    """Use 0.25 for each arm (p=0.5); very conservative."""


@unique
class QuantileMethod(StrEnum):
    LOWER = "Lower"
    HIGHER = "Higher"
    NEAREST = "Nearest"
    LINEAR = "Linear"
    MIDDLE = "Middle"


# @unique
# class RepMethod(StrEnum):
#     JACKKNIFE = "Jackknife"
#     BOOTSTRAP = "Bootstrap"
#     BRR = "BRR"


@unique
class RankScoreMethod(StrEnum):
    """Score function for design-based rank tests.

    KRUSKAL_WALLIS (WILCOXON in teh case of two groups) uses proportional ranks g(r) = r/N.
    VANDER_WAERDEN uses inverse-Normal scores g(r) = Φ⁻¹(r/N).
    MEDIAN uses indicator scores g(r) = I(r > N/2).
    """

    KRUSKAL_WALLIS = "Kruskal-Wallis"
    VANDER_WAERDEN = "vanderWaerden"
    MEDIAN = "Median"


@unique
class SelectMethod(StrEnum):
    SRS_WR = "SRS with replacement"
    SRS_WOR = "SRS without replacement"
    SRS_SYS = "Systematic"
    PPS_BREWER = "PPS Brewer"
    # PPS_HV = "PPS Hanurav-Vijayan"
    PPS_MURPHY = "PPS Murphy"
    PPS_RS = "PPS Rao-Sampford"
    PPS_SYS = "PPS Systematic"
    PPS_WR = "PPS with replacement"
    GRS = "General"


@unique
class SingletonHandling(Enum):
    ERROR = "error"  # Raise error if singletons exist
    CERTAINTY = "certainty"  # PSU→stratum, SSU/records→PSUs
    SKIP = "skip"  # Exclude from variance (data stays)
    COMBINE = "combine"  # Manual column value remapping
    COLLAPSE = "collapse"  # Merge into existing strata
    POOL = "pool"  # Combine singletons into pseudo-stratum
    SCALE = "scale"  # Post-hoc variance inflation (R's "average")
    CENTER = "center"  # Grand-mean centering (R's "adjust") - NotImplemented


# @unique
# class SizeOnePropMethod(StrEnum):
#     WALD = "Wald"
#     FLEISS = "Fleiss"


@unique
class TwoPropsSizeMethod(StrEnum):
    WALD = "Wald (closed-form)"
    NEWCOMBE = "Newcombe"
    MIETTINEN_NURMINEN = "Miettinen-Nurminen"
    FARRINGTON_MANNING = "Farrington-Manning"


@unique
class TableType(StrEnum):
    ONE_WAY = "One-Way"
    TWO_WAY = "Two-Way"


@unique
class TableUnits(StrEnum):
    PROPORTION = "Proportion"
    PERCENT = "Percent"
    COUNT = "Count"


@unique
class TTestType(StrEnum):
    ONE_SAMPLE = "One-Sample"
    TWO_SAMPLE = "Two-Sample"



# src/svy/size/base.py
from __future__ import annotations

import copy
import logging
import re

from collections.abc import Iterable, Mapping
from typing import Self, cast

import msgspec
import numpy as np

from svy.core.enumerations import OnePropSizeMethod, PopParam, PropVarMode, TwoPropsSizeMethod
from svy.core.types import DomainScalarMap, Number
from svy.engine.size_and_power.size import (
    _apply_deff,
    _apply_deff_pair,
    _apply_fpc_srswor,
    _apply_fpc_srswor_pair,
    _apply_nonresponse,
    _apply_nonresponse_pair,
    _fleiss_sample_size_prop,
    _wald_sample_size_mean,
    _wald_sample_size_prop,
    _wald_sample_size_two_props,
)
from svy.utils.helpers import _get_keys_from_maps


log = logging.getLogger(__name__)

_CHUNK_RE = re.compile(r"(\d+)")


# =============================================================================
# Optional central UI integration (svy.ui.printing)
# =============================================================================


def _ui_module():
    """Return svy.ui.printing if available, else None (soft dependency)."""
    try:
        from svy.ui import printing as _p

        return _p
    except Exception:
        return None


def _resolve_width_for(obj, default: int = 88) -> int:
    """
    Ask svy.ui.printing for a preferred width when available; otherwise use a
    robust fallback chain (instance → class → module var → __main__ → builtins
    → env → terminal → default).
    """
    P = _ui_module()
    if P:
        for call in (
            lambda: P.resolve_width(owner=obj, default=default),
            lambda: P.resolve_width(default=default),
            lambda: P.resolve_width(),
        ):
            try:
                w = int(call())
                if w > 20:
                    return w
            except Exception:
                pass

    import builtins
    import os
    import shutil
    import sys

    def _as_int(x):
        try:
            return int(x)
        except Exception:
            return None

    w = _as_int(getattr(obj, "_print_width", None))
    if isinstance(w, int) and w > 20:
        return w

    w = _as_int(getattr(type(obj), "PRINT_WIDTH", None))
    if isinstance(w, int) and w > 20:
        return w

    try:
        mod = sys.modules.get(obj.__class__.__module__)
        if mod is not None and hasattr(mod, "SVY_PRINT_WIDTH"):
            mw = _as_int(getattr(mod, "SVY_PRINT_WIDTH"))
            if isinstance(mw, int) and mw > 20:
                return mw
    except Exception:
        pass

    main_mod = sys.modules.get("__main__")
    if main_mod is not None and hasattr(main_mod, "SVY_PRINT_WIDTH"):
        mw = _as_int(getattr(main_mod, "SVY_PRINT_WIDTH"))
        if isinstance(mw, int) and mw > 20:
            return mw

    bw = _as_int(getattr(builtins, "SVY_PRINT_WIDTH", None))
    if isinstance(bw, int) and bw > 20:
        return bw

    ew = _as_int(os.environ.get("SVY_PRINT_WIDTH"))
    if isinstance(ew, int) and ew > 20:
        return ew

    try:
        ts = shutil.get_terminal_size(fallback=(default, 24))
        if isinstance(ts.columns, int) and ts.columns > 20:
            return ts.columns
    except Exception:
        pass

    return default


def _panel_settings_for(
    obj,
    *,
    title_hint: str = "Sample Size",
    border_fallback: str = "cyan",
    feature_key: str = "sampsize.results",
) -> tuple[bool, str, str]:
    """
    Resolve (enabled, title, border_color) for an outer panel. Uses central UI
    if available; falls back to class vars, then env override.
    """
    show_panel = bool(getattr(type(obj), "_PRINT_PANEL", True))
    title = getattr(type(obj), "_PANEL_TITLE", title_hint)
    border = getattr(type(obj), "_PANEL_BORDER", border_fallback)

    P = _ui_module()
    if P:
        try:
            enabled = P.panel_enabled(feature_key)
            if isinstance(enabled, bool):
                show_panel = enabled
        except Exception:
            pass
        try:
            st = P.styles(feature_key, title=title, border=border)
            if isinstance(st, dict):
                title = st.get("title", title)
                border = st.get("border", border)
        except Exception:
            pass

    import os as _os

    env_panel = _os.environ.get("SVY_PRINT_PANEL")
    if env_panel is not None:
        show_panel = env_panel.lower() not in {"0", "false", "off", "no"}

    return show_panel, title, border


# =============================================================================
# Core data structures
# =============================================================================


class SampleSizeConfigs(msgspec.Struct):
    stratified: bool = False
    pop_size: Number | DomainScalarMap | None = None
    deff: Number | DomainScalarMap = 1.0
    resp_rate: Number | DomainScalarMap = 1.0
    alloc_unit: Number | DomainScalarMap = 1  # round up to multiple (e.g., cluster size)


class TargetProp(msgspec.Struct, frozen=True, tag="prop"):
    p: Number
    moe: Number
    alpha: Number = 0.05


class TargetMean(msgspec.Struct, frozen=True, tag="mean"):
    sigma: Number
    moe: Number
    alpha: Number = 0.05


class TargetTwoProps(msgspec.Struct, frozen=True, tag="two_prop"):
    p1: Number
    p2: Number
    alloc_ratio: Number = 1.0  # allocation ratio n2/n1
    alpha: Number = 0.05
    power: Number = 0.80


class TargetTwoMeans(msgspec.Struct, frozen=True, tag="two_prop"):
    mu1: Number
    mu2: Number
    alloc_ratio: Number = 1.0  # allocation ratio n2/n1
    alpha: Number = 0.05
    power: Number = 0.80


Target = TargetProp | TargetMean | TargetTwoProps | TargetTwoMeans


class Size(msgspec.Struct, frozen=True, tag="size"):
    stratum: str | None = None
    n0: Number | tuple[Number, Number] = 0  # base (no FPC/DEFF/response)
    n1_fpc: Number | tuple[Number, Number] | None = None  # after FPC (if pop_size provided)
    n2_deff: Number | tuple[Number, Number] | None = None  # after DEFF (if response provided)
    n: Number | tuple[Number, Number] = (
        0  # final rounded after nonresponse adjustment; for two-arm: (n1, n2)
    )


class SampleSize:
    """
    Compute required sample sizes for survey objectives under simple or stratified designs.

    Parameters
    ----------
    pop_size : Number | DomainScalarMap | None, default None
        Target population size. When provided as a mapping, values are interpreted
        per stratum (keys are stratum labels). If omitted, no finite population
        correction is applied.
    deff : Number | DomainScalarMap, default 1.0
        Design effect. May be a scalar or a per-stratum mapping.
    resp_rate : Number | DomainScalarMap, default 1.0
        Anticipated unit response rate. May be a scalar or a per-stratum mapping.
    alloc_unit : Number | DomainScalarMap, default 1
        Allocation unit (e.g., cluster/PSU size or minimum take). May be a scalar
        or a per-stratum mapping.
    stratified : bool, default False
        If ``True``, interpret inputs as stratified. See notes below for rules.
    n_strata : int | None, optional
        Number of strata when ``stratified=True`` and all other parameters are
        scalar; values will be replicated across strata.

    Notes
    -----
    - **Unstratified**: all of ``pop_size``, ``deff``, ``resp_rate``, and
      ``alloc_unit`` must be scalars.
    - **Stratified (per-stratum mappings provided)**: pass dicts for any of the
      parameters to supply per-stratum values.
    - **Stratified (all scalars)**: set ``n_strata`` to a positive integer; scalar
      values will be replicated across strata.
    """

    # ---------- printing/branding toggles (class-level fallbacks) ----------
    _PRINT_PANEL: bool = True
    _PANEL_TITLE: str = "Sample Size"
    _PANEL_BORDER: str = "cyan"
    PRINT_WIDTH: int | None = None  # optional class-level width override

    __slots__ = ("_configs", "_param", "_target", "_size")

    def __init__(
        self,
        pop_size: Number | DomainScalarMap | None = None,
        deff: Number | DomainScalarMap = 1.0,
        resp_rate: Number | DomainScalarMap = 1.0,
        alloc_unit: Number | DomainScalarMap = 1,
        stratified: bool = False,
        n_strata: Number | None = None,
    ):
        at_least_one_dict = (
            isinstance(pop_size, dict)
            or isinstance(deff, dict)
            or isinstance(resp_rate, dict)
            or isinstance(alloc_unit, dict)
        )

        if not stratified and at_least_one_dict:
            raise ValueError(
                "pop_size, deff, resp_rate, and alloc_unit must be scalar when stratified=False"
            )
        elif not stratified:
            self._configs = SampleSizeConfigs(
                stratified=False,
                pop_size=pop_size,
                deff=deff,
                resp_rate=resp_rate,
                alloc_unit=alloc_unit,
            )
        else:
            if (n_strata is None or n_strata <= 0) and not at_least_one_dict:
                raise ValueError(
                    "n_strata must be a positive integer when stratified=True and all other parameters are scalar"
                )
            elif at_least_one_dict:
                strata = _get_keys_from_maps(
                    **{
                        k: v
                        for k, v in dict(
                            pop_size=pop_size,
                            deff=deff,
                            resp_rate=resp_rate,
                            alloc_unit=alloc_unit,
                        ).items()
                        if isinstance(v, Mapping)
                    }
                )
                pop_size_map = (
                    {k: pop_size for k in strata} if isinstance(pop_size, Number) else pop_size
                )
                deff_map = {k: deff for k in strata} if isinstance(deff, Number) else deff
                resp_rate_map = (
                    {k: resp_rate for k in strata} if isinstance(resp_rate, Number) else resp_rate
                )
                alloc_unit_map = (
                    {k: alloc_unit for k in strata}
                    if isinstance(alloc_unit, Number)
                    else alloc_unit
                )

                self._configs = SampleSizeConfigs(
                    stratified=True,
                    pop_size=pop_size_map,
                    deff=deff_map,
                    resp_rate=resp_rate_map,
                    alloc_unit=alloc_unit_map,
                )
            else:
                assert n_strata is not None and n_strata > 0
                pop_size_dict = {}
                deff_dict = {}
                resp_rate_dict = {}
                alloc_unit_dict = {}
                for i in range(int(n_strata)):
                    pop_size_dict[i] = pop_size
                    deff_dict[i] = deff
                    resp_rate_dict[i] = resp_rate
                    alloc_unit_dict[i] = alloc_unit
                self._configs = SampleSizeConfigs(
                    stratified=True,
                    pop_size=pop_size_dict,
                    deff=deff_dict,
                    resp_rate=resp_rate_dict,
                    alloc_unit=alloc_unit_dict,
                )

        self._param = None
        self._target = None
        self._size = None

    # ---------- internal helpers ----------
    @staticmethod
    def _is_size_obj(obj) -> bool:
        return all(hasattr(obj, a) for a in ("n0", "n1_fpc", "n2_deff", "n"))

    def _iter_sizes(self):
        """Return a flat list of Size-like objects from self.size/_size."""
        sizes = getattr(self, "size", None)
        if sizes is None:
            sizes = getattr(self, "_size", None)
        if sizes is None:
            return []

        if self._is_size_obj(sizes):  # single Size
            return [sizes]

        if isinstance(sizes, Mapping):  # dict of Size or dict -> [Size]
            out = []
            for v in sizes.values():
                if self._is_size_obj(v):
                    out.append(v)
                elif isinstance(v, Iterable) and not isinstance(v, (str, bytes)):
                    out.extend([x for x in v if self._is_size_obj(x)])
            return out

        if isinstance(sizes, Iterable) and not isinstance(sizes, (str, bytes)):
            return [v for v in sizes if self._is_size_obj(v)]

        return []

    @staticmethod
    def _fmt_num(x) -> str:
        try:
            xf = float(x)
            xi = round(xf)
            return f"{xi:.0f}" if abs(xf - xi) < 1e-9 else f"{xf:.1f}"
        except Exception:
            return str(x)

    @staticmethod
    def _nat_key(label: object):
        """Natural sort: 'region2' < 'region10'; None/'overall' first."""
        if label is None:
            return (0, [""])
        s = str(label).strip().lower()
        is_overall = 0 if s in {"", "overall", "all", "national"} else 1
        parts = _CHUNK_RE.split(s)
        key = [int(p) if p.isdigit() else p for p in parts]
        return (is_overall, key)

    # ---------------- public: Polars export -------------------
    def to_polars(
        self,
        order_by: str | None = None,
        ascending: bool = True,
        natural: bool = True,
        overall_first: bool = True,  # kept for API compatibility
        group_labels: list[str] | None = None,
        include_stratum: bool | None = None,
    ):
        """
        Normalize sizes to a Polars DataFrame.

        - If `include_stratum` is None, it defaults to `self._configs.stratified`.
        - When `include_stratum` is False, the resulting DF will NOT have a 'stratum' column.
        - Handles scalar, stratified, and comparison (tuple) sizes.
        """
        try:
            import polars as pl
        except Exception as e:
            raise ImportError("to_polars() requires the 'polars' package") from e

        if include_stratum is None:
            include_stratum = bool(
                getattr(self, "_configs", None) and getattr(self._configs, "stratified", False)
            )

        objs = self._iter_sizes()
        if not objs:
            cols = (["stratum"] if include_stratum else []) + ["n0", "n1_fpc", "n2_deff", "n"]
            return pl.DataFrame({c: [] for c in cols})

        # detect tuple (comparison) case and validate lengths
        from collections.abc import Iterable as _Iterable

        def _len_if_iter(x):
            return len(x) if isinstance(x, _Iterable) and not isinstance(x, (str, bytes)) else None

        fields = ("n0", "n1_fpc", "n2_deff", "n")
        group_len = None
        for s in objs:
            lens = {f: _len_if_iter(getattr(s, f)) for f in fields}
            ls = [L for L in lens.values() if L is not None]
            if ls:
                L0 = ls[0]
                if any(L != L0 for L in ls):
                    raise ValueError(f"Inconsistent tuple lengths within a Size: {lens}")
                if group_len is None:
                    group_len = L0
                elif group_len != L0:
                    raise ValueError(
                        f"Inconsistent tuple lengths across strata: expected {group_len}, got {L0}"
                    )

        rows = []
        if group_len is None:
            # scalar case
            for s in objs:
                label = getattr(s, "stratum", None) or getattr(s, "domain", None) or "overall"
                row: dict[str, object] = {
                    "n0": float(s.n0),
                    "n1_fpc": float(s.n1_fpc),
                    "n2_deff": float(s.n2_deff),
                    "n": float(s.n),
                }
                if include_stratum:
                    row["stratum"] = "" if label is None else str(label)
                rows.append(row)
            df = pl.DataFrame(rows)
            cols = (["stratum"] if include_stratum else []) + ["n0", "n1_fpc", "n2_deff", "n"]
            df = df.select(cols)
        else:
            # tuple (comparison) case
            if not group_labels:
                group_labels = [f"g{i + 1}" for i in range(group_len)]
            if len(group_labels) != group_len:
                raise ValueError(f"group_labels length {len(group_labels)} != {group_len}")

            for s in objs:
                label = getattr(s, "stratum", None) or getattr(s, "domain", None) or "overall"
                n0, n1, n2, n = (getattr(s, f) for f in fields)
                for i in range(group_len):
                    row: dict[str, object] = {
                        "group": group_labels[i],
                        "n0": float(n0[i]),
                        "n1_fpc": float(n1[i]),
                        "n2_deff": float(n2[i]),
                        "n": float(n[i]),
                    }
                    if include_stratum:
                        row["stratum"] = "" if label is None else str(label)
                    rows.append(row)
            df = pl.DataFrame(rows)
            cols = (["stratum"] if include_stratum else []) + [
                "group",
                "n0",
                "n1_fpc",
                "n2_deff",
                "n",
            ]
            df = df.select(cols)

        # ----- sorting defaults / behavior
        if order_by is None:
            if include_stratum:
                order_by = "stratum"
            else:
                order_by = "group" if ("group" in df.columns) else "n"

        if order_by == "stratum" and include_stratum and "stratum" in df.columns:
            if natural:
                lower = df["stratum"].str.to_lowercase()
                is_overall = lower.is_in(["", "overall", "all", "national"]).cast(pl.Int8)
                # overall first (1 - is_overall) → 0 for overall; reverse if overall_first=False
                overall_key = (1 - is_overall) if overall_first else is_overall
                df = (
                    df.with_columns(
                        [
                            overall_key.alias("_overall"),
                            lower.str.replace_all(r"\d+", "").alias("_prefix"),
                            lower.str.extract(r"(\d+)").cast(pl.Int64).fill_null(0).alias("_num"),
                        ]
                    )
                    .sort(["_overall", "_prefix", "_num"], descending=[not ascending] * 3)
                    .drop(["_overall", "_prefix", "_num"])
                )
            else:
                df = df.sort("stratum", descending=not ascending)

        elif order_by == "group" and "group" in df.columns:
            if natural:
                gl = df["group"].str.to_lowercase()
                df = (
                    df.with_columns(
                        [
                            gl.str.replace_all(r"\d+", "").alias("_gpre"),
                            gl.str.extract(r"(\d+)").cast(pl.Int64).fill_null(0).alias("_gnum"),
                        ]
                    )
                    .sort(["_gpre", "_gnum"], descending=[not ascending, not ascending])
                    .drop(["_gpre", "_gnum"])
                )
            else:
                df = df.sort("group", descending=not ascending)

        elif order_by in {"n", "n0", "n1_fpc", "n2_deff"}:
            df = df.sort(order_by, descending=not ascending)
        else:
            pass

        # Stable secondary sort when both columns exist
        if include_stratum and "group" in df.columns and order_by in {"stratum", "group"}:
            gl = df["group"].str.to_lowercase() if "group" in df.columns else None
            if gl is not None:
                df = (
                    df.with_columns(
                        [
                            df["stratum"]
                            .str.to_lowercase()
                            .str.replace_all(r"\d+", "")
                            .alias("_spre"),
                            df["stratum"]
                            .str.to_lowercase()
                            .str.extract(r"(\d+)")
                            .cast(pl.Int64)
                            .fill_null(0)
                            .alias("_snum"),
                            gl.str.replace_all(r"\d+", "").alias("_gpre2"),
                            gl.str.extract(r"(\d+)").cast(pl.Int64).fill_null(0).alias("_gnum2"),
                        ]
                    )
                    .sort(
                        ["_spre", "_snum", "_gpre2", "_gnum2"],
                        descending=[False, False, False, False],
                    )
                    .drop(["_spre", "_snum", "_gpre2", "_gnum2"])
                )

        return df

    def _format_table_ascii(self) -> str:
        """ASCII fallback that hides 'stratum' when not stratified."""
        objs = self._iter_sizes()
        if not objs:
            return "<no sizes>"

        stratified = bool(
            getattr(self, "_configs", None) and getattr(self._configs, "stratified", False)
        )

        from collections.abc import Iterable as _Iterable

        any_tuple = any(
            isinstance(getattr(s, "n"), _Iterable)
            and not isinstance(getattr(s, "n"), (str, bytes))
            for s in objs
        )

        # Build rows according to shape & stratified flag
        if any_tuple:
            headers = (
                ("group", "n0", "n1_fpc", "n2_deff", "n")
                if not stratified
                else ("stratum", "group", "n0", "n1_fpc", "n2_deff", "n")
            )
            rows = []
            for s in objs:
                label = getattr(s, "stratum", None) or getattr(s, "domain", None) or "overall"
                n0, n1, n2, n = s.n0, s.n1_fpc, s.n2_deff, s.n
                m = len(n)
                for i in range(m):
                    if stratified:
                        rows.append((str(label), f"g{i + 1}", n0[i], n1[i], n2[i], n[i]))
                    else:
                        rows.append((f"g{i + 1}", n0[i], n1[i], n2[i], n[i]))
            # natural sort
            if stratified:
                rows.sort(key=lambda r: (self._nat_key(r[0]), self._nat_key(r[1])))
            else:
                rows.sort(key=lambda r: self._nat_key(r[0]))
        else:
            headers = (
                ("n0", "n1_fpc", "n2_deff", "n")
                if not stratified
                else ("stratum", "n0", "n1_fpc", "n2_deff", "n")
            )
            rows = []
            for s in objs:
                label = getattr(s, "stratum", None) or getattr(s, "domain", None) or "overall"
                if stratified:
                    rows.append((str(label), s.n0, s.n1_fpc, s.n2_deff, s.n))
                else:
                    rows.append((s.n0, s.n1_fpc, s.n2_deff, s.n))

            if stratified:
                rows.sort(key=lambda r: self._nat_key(r[0]))

        # format numbers
        def fmt_row(r):
            # left text cols stay raw; numeric cols formatted
            if len(headers) == 6:  # stratified + groups
                return (r[0], r[1], *(self._fmt_num(v) for v in r[2:]))
            elif len(headers) == 5:  # stratified scalar
                return (r[0], *(self._fmt_num(v) for v in r[1:]))
            elif len(headers) == 4:  # non-stratified scalar
                return tuple(self._fmt_num(v) for v in r)
            else:  # non-stratified groups (5 cols)
                return (r[0], *(self._fmt_num(v) for v in r[1:]))

        formatted = [tuple(map(str, fmt_row(r))) for r in rows]

        # compute widths
        cols = list(zip(*([headers] + formatted)))
        widths = [max(len(c) for c in col) for col in cols]

        def line(cells):
            out = []
            for i, (c, w) in enumerate(zip(cells, widths)):
                left_aligned = headers[i] in {"stratum", "group"}
                out.append(c.ljust(w) if left_aligned else c.rjust(w))
            return "  ".join(out)

        out = [line(headers), line(["-" * w for w in widths])]
        out += [line(r) for r in formatted]
        return "\n".join(out)

    # ---------- indent/surround helper ----------
    @staticmethod
    def _pad_and_surround(text: str, *, indent: int = 2, surround: bool = False) -> str:
        text = str(text).rstrip("\n")
        if indent > 0:
            pad = " " * indent
            text = "\n".join(pad + line if line else pad for line in text.splitlines())
        if surround:
            return f"\n{text}\n"
        return text

    # ----------------- width-aware, panel-styled string -------------------
    def __plain_str__(self) -> str:
        """Plain-text fallback when rich is not installed. Never calls str(self)."""
        ascii_tbl = self._format_table_ascii()
        return self._pad_and_surround(ascii_tbl, indent=2, surround=False)

    def __str__(self) -> str:
        from svy.ui.printing import render_rich_to_str

        return render_rich_to_str(self, width=_resolve_width_for(self))

    def __repr__(self) -> str:
        n = len(self._iter_sizes())
        stratified = getattr(self._configs, "stratified", False)
        param = getattr(self._param, "name", None) or "uncomputed"
        return f"SampleSize(stratified={stratified}, strata={n}, param={param})"

    # ---------------- Rich integration (optional) --------------
    def to_rich_table(self, title: str | None = None):
        """Return a Rich Table (requires `rich`)."""
        try:
            from rich import box
            from rich.table import Table
        except Exception as e:
            raise ImportError("`rich` is required for to_rich_table(). `pip install rich`") from e

        # Try to use Polars for normalized rows
        cols, rows = None, None
        try:
            df = self.to_polars()
            cols, rows = df.columns, df.rows()
            start_idx = 2 if ("group" in cols) else 1
            rows = [
                tuple((v if i < start_idx else self._fmt_num(v)) for i, v in enumerate(row))
                for row in rows
            ]
        except Exception:
            # fallback to ASCII normalization
            from rich.text import Text

            t = Table(title=title, box=box.SIMPLE)
            t.add_column("output")
            t.add_row(Text(self._format_table_ascii()))
            return t

        table = Table(
            title=title, box=box.SIMPLE, show_header=True, header_style="bold", pad_edge=False
        )
        for i, c in enumerate(cols):
            justify = "left" if c in ("stratum", "group") else "right"
            table.add_column(c, justify=justify, overflow="fold")
        for r in rows:
            table.add_row(*[str(x) for x in r])
        return table

    def __rich_console__(self, console, options):
        try:
            from rich.console import Group
            from rich.padding import Padding
            from rich.panel import Panel

            tbl = self.to_rich_table()
            content = Group(Padding(tbl, (0, 0, 0, 2)))

            show_panel, panel_title, border = _panel_settings_for(
                self,
                title_hint=getattr(type(self), "_PANEL_TITLE", "Sample Size"),
                border_fallback=getattr(type(self), "_PANEL_BORDER", "cyan"),
                feature_key="sampsize.results",
            )

            if show_panel:
                yield Panel(content, title=panel_title, border_style=border)
            else:
                yield content
        except Exception:
            from rich.text import Text

            yield Text(self.__plain_str__())

    ## PROPERTY METHODS

    @property
    def configs(self) -> SampleSizeConfigs:
        return self._configs

    @property
    def target(self) -> Target | None:
        return self._target

    @property
    def size(self) -> Size | list[Size] | None:
        return self._size

    @property
    def param(self) -> PopParam | None:
        """Return a defensive copy to avoid external mutation."""
        return copy.copy(self._param)

    ## SAMPLE SIZES FOR PROPORTIONS

    def estimate_prop(
        self,
        p: Number | DomainScalarMap,
        moe: Number | DomainScalarMap,
        *,
        method: OnePropSizeMethod | DomainScalarMap = OnePropSizeMethod.WALD,
        alpha: Number | DomainScalarMap = 0.05,
    ) -> Self:
        self._param = PopParam.PROP

        at_least_one_dict = (
            isinstance(p, dict)
            or isinstance(moe, dict)
            or isinstance(method, dict)
            or isinstance(alpha, dict)
        )

        if self._configs.stratified and not at_least_one_dict:
            raise ValueError(
                "At least one of p, moe, method, or alpha must be a dictionary when stratified sampling is used."
            )
        elif self._configs.stratified:
            strata = _get_keys_from_maps(
                **{
                    k: v
                    for k, v in dict(p=p, moe=moe, method=method, alpha=alpha).items()
                    if isinstance(v, Mapping)
                }
            )
            m = len(strata)
            params = {"p": p, "moe": moe, "method": method, "alpha": alpha}
            for k, v in list(params.items()):
                if isinstance(v, (int, float, np.integer, np.floating)) and not isinstance(
                    v, bool
                ):
                    params[k] = dict(zip(strata, [v] * m))
            p, moe, method, alpha = params["p"], params["moe"], params["method"], params["alpha"]
        else:
            all_scalars = all(isinstance(x, (int, float)) for x in [p, moe, alpha])
            assert all_scalars, "All inputs must be scalars when not stratified."

        match method:
            case OnePropSizeMethod.WALD:
                n0 = _wald_sample_size_prop(
                    target=p,
                    half_ci=moe,
                    alpha=alpha,
                )
            case OnePropSizeMethod.FLEISS:
                n0 = _fleiss_sample_size_prop(
                    target=p,
                    half_ci=moe,
                    alpha=alpha,
                )
            case _:
                raise ValueError(f"Invalid method: {method}")

        if self._configs.pop_size:
            n1_fpc = _apply_fpc_srswor(n0=n0, pop_size=self._configs.pop_size)
        else:
            n1_fpc = n0
        n2_deff = _apply_deff(n=n1_fpc, deff=self._configs.deff)
        n_final = _apply_nonresponse(n=n2_deff, resp_rate=self._configs.resp_rate)

        if not self._configs.stratified:
            self._size = Size(
                stratum=None,
                n0=cast(Number, n0),
                n1_fpc=cast(Number, n1_fpc),
                n2_deff=cast(Number, n2_deff),
                n=cast(Number, n_final),
            )
        else:
            _n0 = cast(DomainScalarMap, n0)
            _n1_fpc = cast(DomainScalarMap, n1_fpc)
            _n2_deff = cast(DomainScalarMap, n2_deff)
            _n_final = cast(DomainScalarMap, n_final)
            self._size = [
                Size(
                    stratum=str(s),
                    n0=_n0[s],
                    n1_fpc=_n1_fpc[s],
                    n2_deff=_n2_deff[s],
                    n=_n_final[s],
                )
                for s in _n0
            ]

        return self

    def compare_props(
        self,
        p1: Number | DomainScalarMap,
        p2: Number | DomainScalarMap,
        *,
        two_sides: bool = True,
        delta: Number | DomainScalarMap = 0.0,
        alloc_ratio: Number = 1.0,  # allocation ratio n2/n1
        method: TwoPropsSizeMethod = TwoPropsSizeMethod.WALD,
        alpha: Number | DomainScalarMap = 0.05,
        power: Number | DomainScalarMap = 0.80,
        var_mode: PropVarMode = PropVarMode.ALT_PROPS,
    ) -> Self:
        self._param = PopParam.PROP

        at_least_one_dict = (
            isinstance(p1, dict)
            or isinstance(p2, dict)
            or isinstance(delta, dict)
            or isinstance(alloc_ratio, dict)
            or isinstance(alpha, dict)
            or isinstance(power, dict)
        )

        if self._configs.stratified and not at_least_one_dict:
            raise ValueError(
                "At least one of p, moe, method, or alpha must be a dictionary when stratified sampling is used."
            )
        elif self._configs.stratified:
            strata = _get_keys_from_maps(
                **{
                    k: v
                    for k, v in dict(
                        p1=p1,
                        p2=p2,
                        delta=delta,
                        alloc_ratio=alloc_ratio,
                        alpha=alpha,
                        power=power,
                    ).items()
                    if isinstance(v, Mapping)
                }
            )
            m = len(strata)
            params = {
                "p1": p1,
                "p2": p2,
                "delta": delta,
                "alloc_ratio": alloc_ratio,
                "alpha": alpha,
                "power": power,
            }
            for k, v in list(params.items()):
                if isinstance(v, (int, float, np.integer, np.floating)) and not isinstance(
                    v, bool
                ):
                    params[k] = dict(zip(strata, [v] * m))

            p1, p2, delta, alloc_ratio, alpha, power = (
                params["p1"],
                params["p2"],
                params["delta"],
                params["alloc_ratio"],
                params["alpha"],
                params["power"],
            )
        else:
            all_scalars = all(
                isinstance(x, (int, float)) for x in [p1, p2, delta, alloc_ratio, alpha, power]
            )
            assert all_scalars, "All inputs must be scalars when not stratified."

        if self._configs.stratified:
            _p1_map = cast(DomainScalarMap, p1)
            _p2_map = cast(DomainScalarMap, p2)
            epsilon: DomainScalarMap = {}
            for s in _p1_map:
                epsilon[s] = _p2_map[s] - _p1_map[s]
        else:
            epsilon = cast(Number, p2) - cast(Number, p1)

        match method:
            case TwoPropsSizeMethod.WALD:
                n0 = _wald_sample_size_two_props(
                    two_sides=two_sides,
                    epsilon=epsilon,
                    prop_1=p1,
                    prop_2=p2,
                    delta=delta,
                    kappa=alloc_ratio,
                    alpha=alpha,
                    power=power,
                    var_mode=var_mode,
                )
            case TwoPropsSizeMethod.MIETTINEN_NURMINEN:
                raise NotImplementedError("MIETTINEN_NURMINEN method is not implemented yet.")
            case TwoPropsSizeMethod.NEWCOMBE:
                raise NotImplementedError("NEWCOMBE method is not implemented yet.")
            case TwoPropsSizeMethod.FARRINGTON_MANNING:
                raise NotImplementedError("FARRINGTON_MANNING method is not implemented yet.")

        if self._configs.pop_size:
            n1_fpc = _apply_fpc_srswor_pair(n0=n0, pop_size=self._configs.pop_size)
        else:
            n1_fpc = n0

        n2_deff = _apply_deff_pair(n=n1_fpc, deff=self._configs.deff)
        n_final = _apply_nonresponse_pair(n=n2_deff, resp_rate=self._configs.resp_rate)

        if not self._configs.stratified:
            self._size = Size(
                stratum=None,
                n0=cast(Number, n0),
                n1_fpc=cast(Number, n1_fpc),
                n2_deff=cast(Number, n2_deff),
                n=cast(Number, n_final),
            )
        else:
            _n0 = cast(DomainScalarMap, n0)
            _n1_fpc = cast(DomainScalarMap, n1_fpc)
            _n2_deff = cast(DomainScalarMap, n2_deff)
            _n_final = cast(DomainScalarMap, n_final)
            self._size = [
                Size(
                    stratum=str(s),
                    n0=_n0[s],
                    n1_fpc=_n1_fpc[s],
                    n2_deff=_n2_deff[s],
                    n=_n_final[s],
                )
                for s in _n0
            ]

        return self

    ## SAMPLE SIZES FOR MEANS

    def estimate_mean(
        self,
        sigma: Number | DomainScalarMap,
        moe: Number | DomainScalarMap,
        *,
        method: OnePropSizeMethod | DomainScalarMap = OnePropSizeMethod.WALD,
        alpha: Number = 0.05,
    ) -> Self:
        self._param = PopParam.MEAN

        at_least_one_dict = (
            isinstance(sigma, dict)
            or isinstance(moe, dict)
            or isinstance(method, dict)
            or isinstance(alpha, dict)
        )

        if self._configs.stratified and not at_least_one_dict:
            raise ValueError(
                "At least one of p, moe, method, or alpha must be a dictionary when stratified sampling is used."
            )
        elif self._configs.stratified:
            strata = _get_keys_from_maps(
                **{
                    k: v
                    for k, v in dict(sigma=sigma, moe=moe, method=method, alpha=alpha).items()
                    if isinstance(v, Mapping)
                }
            )
            m = len(strata)
            params = {"sigma": sigma, "moe": moe, "method": method, "alpha": alpha}
            for k, v in list(params.items()):
                if isinstance(v, (int, float, np.integer, np.floating)) and not isinstance(
                    v, bool
                ):
                    params[k] = dict(zip(strata, [v] * m))

            sigma, moe, method, alpha = (
                params["sigma"],
                params["moe"],
                params["method"],
                params["alpha"],
            )
        else:
            all_scalars = all(isinstance(x, (int, float)) for x in [sigma, moe, alpha])
            assert all_scalars, "All inputs must be scalars when not stratified."

        match method:
            case OnePropSizeMethod.WALD:
                n0 = _wald_sample_size_mean(
                    half_ci=moe,
                    sigma=sigma,
                    alpha=alpha,
                )
            case OnePropSizeMethod.FLEISS:
                raise NotImplementedError("Fleiss method is not implemented for the mean.")
            case _:
                raise ValueError(f"Invalid method: {method}")

        if self._configs.pop_size:
            n1_fpc = _apply_fpc_srswor(n0=n0, pop_size=self._configs.pop_size)
        else:
            n1_fpc = n0

        n2_deff = _apply_deff(n=n1_fpc, deff=self._configs.deff)
        n_final = _apply_nonresponse(n=n2_deff, resp_rate=self._configs.resp_rate)

        if not self._configs.stratified:
            self._size = Size(
                stratum=None,
                n0=cast(Number, n0),
                n1_fpc=cast(Number, n1_fpc),
                n2_deff=cast(Number, n2_deff),
                n=cast(Number, n_final),
            )
        else:
            _n0 = cast(DomainScalarMap, n0)
            _n1_fpc = cast(DomainScalarMap, n1_fpc)
            _n2_deff = cast(DomainScalarMap, n2_deff)
            _n_final = cast(DomainScalarMap, n_final)
            self._size = [
                Size(
                    stratum=str(s),
                    n0=_n0[s],
                    n1_fpc=_n1_fpc[s],
                    n2_deff=_n2_deff[s],
                    n=_n_final[s],
                )
                for s in _n0
            ]

        return self

    def compare_means(
        self,
        mu1: Number,
        mu2: Number,
        *,
        method: OnePropSizeMethod | DomainScalarMap = OnePropSizeMethod.WALD,
        alloc_ratio: Number = 1.0,  # allocation ratio n2/n1
        alpha: Number = 0.05,
        power: Number = 0.80,
    ) -> Self:
        self._param = PopParam.MEAN
        # placeholder for future implementation; kept for API symmetry
        return self



# src/svy/core/terms.py
from __future__ import annotations

from dataclasses import dataclass
from typing import Union


# A "Feature" can be a simple string (column name) or a Term object
Feature = Union[str, "Term"]


class Term:
    """Base class for all model terms."""

    pass


@dataclass(frozen=True)
class Cat(Term):
    """
    Explicitly treat a variable as categorical.

    Args:
        name: The column name in the dataframe.
        ref: (Optional) A specific level to drop (reference level).
             If None, all levels are usually kept (typical for calibration/raking).
    """

    name: str
    ref: str | int | float | None = None

    def __repr__(self) -> str:
        ref_str = f", ref={self.ref!r}" if self.ref is not None else ""
        return f"Cat('{self.name}'{ref_str})"


@dataclass(frozen=True)
class Cross(Term):
    """
    Create an interaction (crossing) between two variables.

    Args:
        left: The first variable (string name or Term object).
        right: The second variable (string name or Term object).
    """

    left: Feature
    right: Feature

    def __repr__(self) -> str:
        return f"Cross({self.left!r}, {self.right!r})"


@dataclass(frozen=True)
class RE(Term):
    """
    Random Effect term.
    Used to define the nesting structure (Area) for SAE.
    """

    name: str

    def __repr__(self) -> str:
        return f"RE('{self.name}')"
